@@ -29,7 +29,7 @@ pid=$!
 trap 'kill ${pid}' INT TERM
 
 set -ex
-"${CARGO:-cargo}" test "${@}" 2>&1 | tee syd.log
+${CARGO:-cargo} test "${@}" 2>&1 | tee syd.log
 r=${PIPESTATUS[0]}
 kill ${pid}
 exit $r
