@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/hash.rs: Utilities for caching
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -21,7 +21,7 @@ use nix::{errno::Errno, sys::socket::UnixAddr, unistd::Pid};
 use serde::{ser::SerializeMap, Serializer};
 
 use crate::{
-    confine::{ScmpNotifReq, SydArch, SydMemoryMap},
+    confine::{ScmpNotifReq, SydArch},
     elf::ExecutableFile,
     hash::{SydHashMap, SydHashSet},
     sigset::SydSigSet,
@@ -69,12 +69,6 @@ pub(crate) struct SysInterruptMap {
 pub(crate) struct ExecResult {
     pub(crate) exe: ExecutableFile,
     pub(crate) file: File,
-    pub(crate) ip: u64,
-    pub(crate) sp: u64,
-    pub(crate) args: [u64; 6],
-    pub(crate) ip_mem: Option<[u8; 64]>,
-    pub(crate) sp_mem: Option<[u8; 64]>,
-    pub(crate) memmap: Option<Vec<SydMemoryMap>>,
 }
 
 /// Syscall-agnostic error map.
@@ -174,7 +168,7 @@ impl serde::Serialize for SysInterrupt {
     where
         S: Serializer,
     {
-        let mut map = serializer.serialize_map(Some(3))?;
+        let mut map = serializer.serialize_map(Some(6))?;
 
         let data = &self.request.data;
         let syscall = ScmpSyscall::get_name_by_arch(data.syscall, data.arch)
