@@ -6,10 +6,6 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
-// We allow this to easily write portable code.
-// FIXME: Do not be lazy.
-#![allow(dead_code)]
-
 use std::{
     ffi::CStr,
     mem::MaybeUninit,
@@ -19,7 +15,13 @@ use std::{
     sync::LazyLock,
 };
 
-use libc::{c_uint, c_void};
+use libc::{
+    accept4, c_char, c_int, c_long, c_uint, c_void, dev_t, gid_t, mode_t, off64_t, off_t, shutdown,
+    sockaddr, socket, socketpair, socklen_t, syscall, uid_t, utimensat, SYS_close_range,
+    SYS_execveat, SYS_faccessat2, SYS_fchdir, SYS_fchmodat, SYS_fchownat, SYS_linkat,
+    SYS_memfd_create, SYS_mkdirat, SYS_mknodat, SYS_openat2, SYS_pipe2, SYS_renameat2,
+    SYS_symlinkat, SYS_umask, SYS_uname, SYS_unlinkat, SHUT_RD, SHUT_RDWR, SHUT_WR,
+};
 use libseccomp::ScmpSyscall;
 use nix::{
     errno::Errno,
@@ -27,14 +29,16 @@ use nix::{
     sys::{
         mman::{mmap_anonymous, MapFlags, ProtFlags},
         socket::{bind, connect, AddressFamily, Shutdown, SockFlag, SockType, SockaddrLike},
-        stat::Mode,
+        stat::{Mode, SFlag},
     },
-    unistd::UnlinkatFlags,
+    unistd::{AccessFlags, Gid, Uid, UnlinkatFlags},
     NixPath,
 };
 
 use crate::{
     compat::{set_vma_anon_name, MFdFlags, OpenHow, RenameFlags},
+    fd::AT_EXECVE_CHECK,
+    path::{empty_argv, empty_envp, empty_path},
     rng::fillrandom,
     sealbox::{getpagesize, mprotect_none, mprotect_readonly, mseal},
     uts::UtsName,
@@ -56,13 +60,27 @@ pub enum CookieIdx {
     CloseRangeArg3 = 0,
     CloseRangeArg4,
     CloseRangeArg5,
+    ExecveatArg5,
     Openat2Arg4,
     Openat2Arg5,
+    Faccessat2Arg4,
+    Faccessat2Arg5,
     FchdirArg1,
     FchdirArg2,
     FchdirArg3,
     FchdirArg4,
     FchdirArg5,
+    FchownatArg5,
+    FchmodatArg3,
+    FchmodatArg4,
+    FchmodatArg5,
+    Fchmodat2Arg4,
+    Fchmodat2Arg5,
+    MkdiratArg3,
+    MkdiratArg4,
+    MkdiratArg5,
+    MknodatArg4,
+    MknodatArg5,
     SocketArg3,
     SocketArg4,
     SocketArg5,
@@ -113,6 +131,9 @@ pub enum CookieIdx {
     UnlinkatArg4,
     UnlinkatArg5,
     LinkatArg5,
+    SymlinkatArg3,
+    SymlinkatArg4,
+    SymlinkatArg5,
     Pipe2Arg2,
     Pipe2Arg3,
     Pipe2Arg4,
@@ -259,8 +280,8 @@ pub static SYSCOOKIE_POOL: LazyLock<SyscookiePool> =
 pub(crate) fn safe_close_range(first: c_uint, last: c_uint, flags: c_uint) -> Result<(), Errno> {
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_close_range,
+        syscall(
+            SYS_close_range,
             first,
             last,
             flags,
@@ -279,33 +300,33 @@ pub(crate) fn safe_openat2<Fd: AsFd, P: NixPath + ?Sized>(
     path: &P,
     mut how: OpenHow,
 ) -> Result<OwnedFd, Errno> {
-    // SAFETY: In libc we trust.
-    #[expect(clippy::cast_possible_truncation)]
-    let fd = path.with_nix_path(|cstr| unsafe {
-        libc::syscall(
-            libc::SYS_openat2,
-            dirfd.as_fd().as_raw_fd(),
-            cstr.as_ptr(),
-            std::ptr::addr_of_mut!(how),
-            size_of::<OpenHow>(),
-            SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg4),
-            SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg5),
-        )
-    })? as RawFd;
-
-    Errno::result(fd)?;
+    let res = path.with_nix_path(|cstr| {
+        // SAFETY: In libc we trust.
+        unsafe {
+            syscall(
+                SYS_openat2,
+                dirfd.as_fd().as_raw_fd(),
+                cstr.as_ptr(),
+                std::ptr::addr_of_mut!(how),
+                size_of::<OpenHow>(),
+                SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg4),
+                SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg5),
+            )
+        }
+    })?;
 
     // SAFETY:
     //
-    // `openat2(2)` should return a valid owned fd on success
-    Ok(unsafe { OwnedFd::from_raw_fd(fd) })
+    // `openat2(2)` should return a valid owned fd on success.
+    #[expect(clippy::cast_possible_truncation)]
+    Errno::result(res).map(|r| unsafe { OwnedFd::from_raw_fd(r as RawFd) })
 }
 
 /// socket(2) may be multiplexed by socketcall(2).
-pub static SYS_SOCKET: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+pub static SYS_SOCKET: LazyLock<Option<c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("socket")
         .map(i32::from)
-        .map(libc::c_long::from)
+        .map(c_long::from)
         .ok()
     {
         Some(n) if n < 0 => None,
@@ -316,16 +337,12 @@ pub static SYS_SOCKET: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 
 /// Safe socket(2) confined by syscall cookies.
 #[inline(always)]
-pub fn safe_socket(
-    domain: libc::c_int,
-    stype: libc::c_int,
-    proto: libc::c_int,
-) -> Result<OwnedFd, Errno> {
+pub fn safe_socket(domain: c_int, stype: c_int, proto: c_int) -> Result<OwnedFd, Errno> {
     if let Some(sys_socket) = *SYS_SOCKET {
         // SAFETY: In libc we trust.
         #[expect(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_socket,
                 domain,
                 stype,
@@ -340,7 +357,7 @@ pub fn safe_socket(
         // SAFETY:
         // socketcall(2) on multiplexed architecture.
         // We use libc version for convenience.
-        Errno::result(unsafe { libc::socket(domain, stype, proto) })
+        Errno::result(unsafe { socket(domain, stype, proto) })
     }
     .map(|fd| {
         // SAFETY: socket returns a valid FD on success.
@@ -349,28 +366,29 @@ pub fn safe_socket(
 }
 
 /// socketpair(2) may be multiplexed by socketcall(2).
-pub static SYS_SOCKETPAIR: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
-    match ScmpSyscall::from_name("socketpair")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+pub static SYS_SOCKETPAIR: LazyLock<Option<c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("socketpair")
+            .map(i32::from)
+            .map(c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// Safe socketpair(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_socketpair(
     domain: AddressFamily,
     ty: SockType,
-    proto: libc::c_int,
+    proto: c_int,
     flags: SockFlag,
 ) -> Result<(OwnedFd, OwnedFd), Errno> {
     // Merge SockType and SockFlags together.
-    let mut ty = ty as libc::c_int;
+    let mut ty = ty as c_int;
     ty |= flags.bits();
 
     let mut fds = [-1, -1];
@@ -378,9 +396,9 @@ pub fn safe_socketpair(
     if let Some(sys_socketpair) = *SYS_SOCKETPAIR {
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_socketpair,
-                domain as libc::c_int,
+                domain as c_int,
                 ty,
                 proto,
                 fds.as_mut_ptr(),
@@ -392,9 +410,7 @@ pub fn safe_socketpair(
         // SAFETY:
         // socketcall(2) on multiplexed architecture.
         // We use libc version for convenience.
-        Errno::result(unsafe {
-            libc::socketpair(domain as libc::c_int, ty, proto, fds.as_mut_ptr())
-        })?;
+        Errno::result(unsafe { socketpair(domain as c_int, ty, proto, fds.as_mut_ptr()) })?;
     }
 
     // SAFETY: socketpair returns valid FDs on success.
@@ -402,17 +418,18 @@ pub fn safe_socketpair(
 }
 
 /// accept4(2) may be multiplexed by socketcall(2).
-pub static SYS_ACCEPT4: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
-    match ScmpSyscall::from_name("accept4")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+pub static SYS_ACCEPT4: LazyLock<Option<c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("accept4")
+            .map(i32::from)
+            .map(c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// Safe accept4(2) confined by syscall cookies.
 ///
@@ -422,15 +439,15 @@ pub static SYS_ACCEPT4: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 #[inline(always)]
 pub unsafe fn safe_accept4<Fd: AsFd>(
     fd: Fd,
-    addr: *mut libc::sockaddr,
-    len: *mut libc::socklen_t,
+    addr: *mut sockaddr,
+    len: *mut socklen_t,
     flags: SockFlag,
 ) -> Result<OwnedFd, Errno> {
     if let Some(sys_accept4) = *SYS_ACCEPT4 {
         // SAFETY: In libc we trust.
         #[expect(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_accept4,
                 fd.as_fd().as_raw_fd(),
                 addr,
@@ -446,11 +463,12 @@ pub unsafe fn safe_accept4<Fd: AsFd>(
         })
     } else {
         // SAFETY: socketcall(2) on multiplexed architecture.
-        Errno::result(unsafe { libc::accept4(fd.as_fd().as_raw_fd(), addr, len, flags.bits()) })
-            .map(|fd| {
+        Errno::result(unsafe { accept4(fd.as_fd().as_raw_fd(), addr, len, flags.bits()) }).map(
+            |fd| {
                 // SAFETY: accept4 returns a valid fd on success.
                 unsafe { OwnedFd::from_raw_fd(fd) }
-            })
+            },
+        )
     }
 }
 
@@ -465,10 +483,10 @@ pub fn safer_accept4<Fd: AsFd>(fd: Fd, flags: SockFlag) -> Result<OwnedFd, Errno
 }
 
 /// bind(2) may be multiplexed by socketcall(2).
-pub(crate) static SYS_BIND: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+pub(crate) static SYS_BIND: LazyLock<Option<c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("bind")
         .map(i32::from)
-        .map(libc::c_long::from)
+        .map(c_long::from)
         .ok()
     {
         Some(n) if n < 0 => None,
@@ -483,7 +501,7 @@ pub fn safe_bind<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno>
     if let Some(sys_bind) = *SYS_BIND {
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_bind,
                 fd.as_fd().as_raw_fd(),
                 addr.as_ptr(),
@@ -501,17 +519,18 @@ pub fn safe_bind<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno>
 }
 
 /// connect(2) may be multiplexed by socketcall(2).
-pub static SYS_CONNECT: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
-    match ScmpSyscall::from_name("connect")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+pub static SYS_CONNECT: LazyLock<Option<c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("connect")
+            .map(i32::from)
+            .map(c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// Safe connect(2) confined by syscall cookies.
 #[inline(always)]
@@ -519,7 +538,7 @@ pub fn safe_connect<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Err
     if let Some(sys_connect) = *SYS_CONNECT {
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_connect,
                 fd.as_fd().as_raw_fd(),
                 addr.as_ptr(),
@@ -539,31 +558,32 @@ pub fn safe_connect<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Err
 /// shutdown(2) may be multiplexed by socketcall(2).
 ///
 /// This is used by syd-tor(1) only, syd(1) does not hook into shutdown(2).
-pub static SYS_SHUTDOWN: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
-    match ScmpSyscall::from_name("shutdown")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+pub static SYS_SHUTDOWN: LazyLock<Option<c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("shutdown")
+            .map(i32::from)
+            .map(c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// Safe shutdown(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_shutdown<Fd: AsFd>(fd: Fd, how: Shutdown) -> Result<(), Errno> {
     let how = match how {
-        Shutdown::Read => libc::SHUT_RD,
-        Shutdown::Write => libc::SHUT_WR,
-        Shutdown::Both => libc::SHUT_RDWR,
+        Shutdown::Read => SHUT_RD,
+        Shutdown::Write => SHUT_WR,
+        Shutdown::Both => SHUT_RDWR,
     };
 
     if let Some(sys_shutdown) = *SYS_SHUTDOWN {
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_shutdown,
                 fd.as_fd().as_raw_fd(),
                 how,
@@ -576,7 +596,7 @@ pub fn safe_shutdown<Fd: AsFd>(fd: Fd, how: Shutdown) -> Result<(), Errno> {
         .map(drop)
     } else {
         // SAFETY: socketcall(2) on multiplexed architecture.
-        Errno::result(unsafe { libc::shutdown(fd.as_fd().as_raw_fd(), how) }).map(drop)
+        Errno::result(unsafe { shutdown(fd.as_fd().as_raw_fd(), how) }).map(drop)
     }
 }
 
@@ -591,8 +611,8 @@ pub fn safe_memfd_create<P: NixPath + ?Sized>(name: &P, flags: MFdFlags) -> Resu
     let res = name.with_nix_path(|cstr| {
         // SAFETY: In libc we trust.
         unsafe {
-            libc::syscall(
-                libc::SYS_memfd_create,
+            syscall(
+                SYS_memfd_create,
                 cstr.as_ptr(),
                 flags.bits(),
                 SYSCOOKIE_POOL.get(CookieIdx::MemfdCreateArg2),
@@ -603,13 +623,11 @@ pub fn safe_memfd_create<P: NixPath + ?Sized>(name: &P, flags: MFdFlags) -> Resu
         }
     })?;
 
-    #[expect(clippy::cast_possible_truncation)]
-    let fd = Errno::result(res).map(|fd| fd as RawFd)?;
-
     // SAFETY:
     //
     // `memfd_create(2)` returns a valid owned fd on success.
-    Ok(unsafe { OwnedFd::from_raw_fd(fd) })
+    #[expect(clippy::cast_possible_truncation)]
+    Errno::result(res).map(|r| unsafe { OwnedFd::from_raw_fd(r as RawFd) })
 }
 
 /// Safe renameat2(2) confined by syscall cookies.
@@ -622,42 +640,54 @@ pub(crate) fn safe_renameat2<Fd1: AsFd, Fd2: AsFd, P1: NixPath + ?Sized, P2: Nix
     flags: RenameFlags,
 ) -> Result<(), Errno> {
     let res = old_path.with_nix_path(|old_cstr| {
-        // SAFETY: In libc we trust.
-        new_path.with_nix_path(|new_cstr| unsafe {
-            libc::syscall(
-                libc::SYS_renameat2,
-                old_dirfd.as_fd().as_raw_fd(),
-                old_cstr.as_ptr(),
-                new_dirfd.as_fd().as_raw_fd(),
-                new_cstr.as_ptr(),
-                flags.bits(),
-                SYSCOOKIE_POOL.get(CookieIdx::Renameat2Arg5),
-            )
+        new_path.with_nix_path(|new_cstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    SYS_renameat2,
+                    old_dirfd.as_fd().as_raw_fd(),
+                    old_cstr.as_ptr(),
+                    new_dirfd.as_fd().as_raw_fd(),
+                    new_cstr.as_ptr(),
+                    flags.bits(),
+                    SYSCOOKIE_POOL.get(CookieIdx::Renameat2Arg5),
+                )
+            }
         })
     })??;
-
     Errno::result(res).map(drop)
 }
 
+// Note fchmodat2 may not be available,
+// and libc::SYS_fchmodat2 may not be defined.
+// Therefore we query the number using libseccomp.
+static SYS_FCHMODAT2: LazyLock<c_long> = LazyLock::new(|| {
+    ScmpSyscall::from_name("fchmodat2")
+        .map(i32::from)
+        .map(c_long::from)
+        .unwrap_or(-1) // Invalid system call.
+});
+
 /// truncate(2) may be aliased to truncate64(2) by libc.
-static SYS_TRUNCATE: LazyLock<Option<libc::c_long>> =
-    LazyLock::new(|| {
-        match ScmpSyscall::from_name("truncate")
-            .map(i32::from)
-            .map(libc::c_long::from)
-            .ok()
-        {
-            Some(n) if n < 0 => None,
-            Some(n) => Some(n),
-            None => None,
-        }
-    });
+static SYS_TRUNCATE: LazyLock<Option<c_long>> = LazyLock::new(|| {
+    match ScmpSyscall::from_name("truncate")
+        .map(i32::from)
+        .map(c_long::from)
+        .ok()
+    {
+        Some(n) if n < 0 => None,
+        Some(n) => Some(n),
+        None => None,
+    }
+});
 
 /// truncate64(2) may not always be available via libc.
-static SYS_TRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+// This is not present on some architectures.
+#[allow(dead_code)]
+static SYS_TRUNCATE64: LazyLock<Option<c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("truncate64")
         .map(i32::from)
-        .map(libc::c_long::from)
+        .map(c_long::from)
         .ok()
     {
         Some(n) if n < 0 => None,
@@ -667,10 +697,10 @@ static SYS_TRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 });
 
 /// ftruncate(2) may be aliased to ftruncate64(2) by libc.
-static SYS_FTRUNCATE: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+static SYS_FTRUNCATE: LazyLock<Option<c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("ftruncate")
         .map(i32::from)
-        .map(libc::c_long::from)
+        .map(c_long::from)
         .ok()
     {
         Some(n) if n < 0 => None,
@@ -680,48 +710,48 @@ static SYS_FTRUNCATE: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 });
 
 /// ftruncate64(2) may not always be available via libc.
-static SYS_FTRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
-    match ScmpSyscall::from_name("ftruncate64")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+// This is not present on some architectures.
+#[allow(dead_code)]
+static SYS_FTRUNCATE64: LazyLock<Option<c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("ftruncate64")
+            .map(i32::from)
+            .map(c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// Safe truncate(2) confined by syscall cookies.
-pub(crate) fn safe_truncate<P: NixPath + ?Sized>(path: &P, len: libc::off_t) -> Result<(), Errno> {
+pub(crate) fn safe_truncate<P: NixPath + ?Sized>(path: &P, len: off_t) -> Result<(), Errno> {
     // On ILP32 where off_t is 64-bit, dispatch to truncate64 ABI.
-    if size_of::<libc::off_t>() > size_of::<libc::c_long>() {
-        return safe_truncate64(path, libc::off64_t::from(len));
+    if size_of::<off_t>() > size_of::<c_long>() {
+        return safe_truncate64(path, off64_t::from(len));
     }
 
     let sys_truncate = SYS_TRUNCATE.ok_or(Errno::ENOSYS)?;
-
-    // SAFETY: In libc we trust.
-    let res = path.with_nix_path(|cstr| unsafe {
-        libc::syscall(
-            sys_truncate,
-            cstr.as_ptr(),
-            len,
-            SYSCOOKIE_POOL.get(CookieIdx::TruncateArg2),
-            SYSCOOKIE_POOL.get(CookieIdx::TruncateArg3),
-            SYSCOOKIE_POOL.get(CookieIdx::TruncateArg4),
-            SYSCOOKIE_POOL.get(CookieIdx::TruncateArg5),
-        )
+    let res = path.with_nix_path(|cstr| {
+        // SAFETY: In libc we trust.
+        unsafe {
+            syscall(
+                sys_truncate,
+                cstr.as_ptr(),
+                len,
+                SYSCOOKIE_POOL.get(CookieIdx::TruncateArg2),
+                SYSCOOKIE_POOL.get(CookieIdx::TruncateArg3),
+                SYSCOOKIE_POOL.get(CookieIdx::TruncateArg4),
+                SYSCOOKIE_POOL.get(CookieIdx::TruncateArg5),
+            )
+        }
     })?;
-
     Errno::result(res).map(drop)
 }
 
 /// Safe truncate64(2) confined by syscall cookies.
-pub(crate) fn safe_truncate64<P: NixPath + ?Sized>(
-    path: &P,
-    len: libc::off64_t,
-) -> Result<(), Errno> {
+pub(crate) fn safe_truncate64<P: NixPath + ?Sized>(path: &P, len: off64_t) -> Result<(), Errno> {
     #[cfg(not(any(
         target_pointer_width = "64",
         all(target_arch = "x86_64", target_pointer_width = "32"),
@@ -750,22 +780,24 @@ pub(crate) fn safe_truncate64<P: NixPath + ?Sized>(
 
         // i386: low, high
         let val = len as u64;
-        let low = (val & 0xFFFF_FFFF) as libc::c_long;
-        let high = (val >> 32) as libc::c_long;
+        let low = (val & 0xFFFF_FFFF) as c_long;
+        let high = (val >> 32) as c_long;
 
-        // SAFETY: In libc we trust.
-        Errno::result(path.with_nix_path(|cstr| unsafe {
-            libc::syscall(
-                sys_truncate64,
-                cstr.as_ptr(),
-                low,
-                high,
-                SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg3),
-                SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg4),
-                SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg5),
-            )
-        })?)
-        .map(drop)
+        let res = path.with_nix_path(|cstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    sys_truncate64,
+                    cstr.as_ptr(),
+                    low,
+                    high,
+                    SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg3),
+                    SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg4),
+                    SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg5),
+                )
+            }
+        })?;
+        Errno::result(res).map(drop)
     }
 
     #[cfg(any(
@@ -779,37 +811,39 @@ pub(crate) fn safe_truncate64<P: NixPath + ?Sized>(
 
         // 32-bit ARM/ppc/mips: 0, low, high
         let val = len as u64;
-        let low = (val & 0xFFFF_FFFF) as libc::c_long;
-        let high = (val >> 32) as libc::c_long;
+        let low = (val & 0xFFFF_FFFF) as c_long;
+        let high = (val >> 32) as c_long;
 
-        // SAFETY: In libc we trust.
-        Errno::result(path.with_nix_path(|cstr| unsafe {
-            libc::syscall(
-                sys_truncate64,
-                cstr.as_ptr(),
-                0 as libc::c_long,
-                low,
-                high,
-                SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg4),
-                SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg5),
-            )
-        })?)
-        .map(drop)
+        let res = path.with_nix_path(|cstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    sys_truncate64,
+                    cstr.as_ptr(),
+                    0 as c_long,
+                    low,
+                    high,
+                    SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg4),
+                    SYSCOOKIE_POOL.get(CookieIdx::Truncate64Arg5),
+                )
+            }
+        })?;
+        Errno::result(res).map(drop)
     }
 }
 
 /// Safe ftruncate(2) confined by syscall cookies.
-pub(crate) fn safe_ftruncate<Fd: AsFd>(fd: Fd, len: libc::off_t) -> Result<(), Errno> {
+pub(crate) fn safe_ftruncate<Fd: AsFd>(fd: Fd, len: off_t) -> Result<(), Errno> {
     // On ILP32 where off_t is 64-bit, dispatch to ftruncate64 ABI.
-    if size_of::<libc::off_t>() > size_of::<libc::c_long>() {
-        return safe_ftruncate64(fd, libc::off64_t::from(len));
+    if size_of::<off_t>() > size_of::<c_long>() {
+        return safe_ftruncate64(fd, off64_t::from(len));
     }
 
     let sys_ftruncate = SYS_FTRUNCATE.ok_or(Errno::ENOSYS)?;
 
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::syscall(
+        syscall(
             sys_ftruncate,
             fd.as_fd().as_raw_fd(),
             len,
@@ -823,7 +857,7 @@ pub(crate) fn safe_ftruncate<Fd: AsFd>(fd: Fd, len: libc::off_t) -> Result<(), E
 }
 
 /// Safe ftruncate64(2) confined by syscall cookies.
-pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: libc::off64_t) -> Result<(), Errno> {
+pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: off64_t) -> Result<(), Errno> {
     #[cfg(not(any(
         target_pointer_width = "64",
         all(target_arch = "x86_64", target_pointer_width = "32"),
@@ -852,12 +886,12 @@ pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: libc::off64_t) -> Result<(
 
         // i386: low, high
         let val = len as u64;
-        let low = (val & 0xFFFF_FFFF) as libc::c_long;
-        let high = (val >> 32) as libc::c_long;
+        let low = (val & 0xFFFF_FFFF) as c_long;
+        let high = (val >> 32) as c_long;
 
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_ftruncate64,
                 fd.as_fd().as_raw_fd(),
                 low,
@@ -881,15 +915,15 @@ pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: libc::off64_t) -> Result<(
 
         // 32-bit ARM/ppc/mips: 0, low, high
         let val = len as u64;
-        let low = (val & 0xFFFF_FFFF) as libc::c_long;
-        let high = (val >> 32) as libc::c_long;
+        let low = (val & 0xFFFF_FFFF) as c_long;
+        let high = (val >> 32) as c_long;
 
         // SAFETY: In libc we trust.
         Errno::result(unsafe {
-            libc::syscall(
+            syscall(
                 sys_ftruncate64,
                 fd.as_fd().as_raw_fd(),
-                0 as libc::c_long,
+                0 as c_long,
                 low,
                 high,
                 SYSCOOKIE_POOL.get(CookieIdx::Ftruncate64Arg4),
@@ -912,19 +946,20 @@ pub(crate) fn safe_unlinkat<Fd: AsFd, P: NixPath + ?Sized>(
         UnlinkatFlags::NoRemoveDir => AtFlags::empty(),
     };
 
-    // SAFETY: In libc we trust.
-    let res = path.with_nix_path(|cstr| unsafe {
-        libc::syscall(
-            libc::SYS_unlinkat,
-            dirfd.as_fd().as_raw_fd(),
-            cstr.as_ptr(),
-            atflag.bits(),
-            SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg3),
-            SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg4),
-            SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg5),
-        )
+    let res = path.with_nix_path(|cstr| {
+        // SAFETY: In libc we trust.
+        unsafe {
+            syscall(
+                SYS_unlinkat,
+                dirfd.as_fd().as_raw_fd(),
+                cstr.as_ptr(),
+                atflag.bits(),
+                SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg3),
+                SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg4),
+                SYSCOOKIE_POOL.get(CookieIdx::UnlinkatArg5),
+            )
+        }
     })?;
-
     Errno::result(res).map(drop)
 }
 
@@ -938,30 +973,145 @@ pub(crate) fn safe_linkat<Fd1: AsFd, Fd2: AsFd, P1: NixPath + ?Sized, P2: NixPat
     flag: AtFlags,
 ) -> Result<(), Errno> {
     let res = oldpath.with_nix_path(|oldcstr| {
-        // SAFETY: In libc we trust.
-        newpath.with_nix_path(|newcstr| unsafe {
-            libc::syscall(
-                libc::SYS_linkat,
-                olddirfd.as_fd().as_raw_fd(),
-                oldcstr.as_ptr(),
-                newdirfd.as_fd().as_raw_fd(),
-                newcstr.as_ptr(),
-                flag.bits(),
-                SYSCOOKIE_POOL.get(CookieIdx::LinkatArg5),
-            )
+        newpath.with_nix_path(|newcstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    SYS_linkat,
+                    olddirfd.as_fd().as_raw_fd(),
+                    oldcstr.as_ptr(),
+                    newdirfd.as_fd().as_raw_fd(),
+                    newcstr.as_ptr(),
+                    flag.bits(),
+                    SYSCOOKIE_POOL.get(CookieIdx::LinkatArg5),
+                )
+            }
         })
     })??;
-
     Errno::result(res).map(drop)
 }
 
+/// Safe symlinkat(2) confined by syscall cookies.
+#[inline(always)]
+pub(crate) fn safe_symlinkat<Fd: AsFd, P1: NixPath + ?Sized, P2: NixPath + ?Sized>(
+    path1: &P1,
+    dirfd: Fd,
+    path2: &P2,
+) -> Result<(), Errno> {
+    let res = path1.with_nix_path(|path1| {
+        path2.with_nix_path(|path2| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    SYS_symlinkat,
+                    path1.as_ptr(),
+                    dirfd.as_fd().as_raw_fd(),
+                    path2.as_ptr(),
+                    SYSCOOKIE_POOL.get(CookieIdx::SymlinkatArg3),
+                    SYSCOOKIE_POOL.get(CookieIdx::SymlinkatArg4),
+                    SYSCOOKIE_POOL.get(CookieIdx::SymlinkatArg5),
+                )
+            }
+        })
+    })??;
+    Errno::result(res).map(drop)
+}
+
+/// Safe mkdirat(2) confined by syscall cookies.
+#[inline(always)]
+pub(crate) fn safe_mkdirat<Fd: AsFd, P: NixPath + ?Sized>(
+    dirfd: Fd,
+    path: &P,
+    mode: Mode,
+) -> Result<(), Errno> {
+    let res = path.with_nix_path(|cstr| {
+        // SAFETY: In libc we trust.
+        unsafe {
+            syscall(
+                SYS_mkdirat,
+                dirfd.as_fd().as_raw_fd(),
+                cstr.as_ptr(),
+                mode.bits(),
+                SYSCOOKIE_POOL.get(CookieIdx::MkdiratArg3),
+                SYSCOOKIE_POOL.get(CookieIdx::MkdiratArg4),
+                SYSCOOKIE_POOL.get(CookieIdx::MkdiratArg5),
+            )
+        }
+    })?;
+    Errno::result(res).map(drop)
+}
+
+/// Safe mknodat(2) confined by syscall cookies.
+#[inline(always)]
+pub(crate) fn safe_mknodat<Fd: AsFd, P: NixPath + ?Sized>(
+    dirfd: Fd,
+    path: &P,
+    kind: SFlag,
+    perm: Mode,
+    dev: dev_t,
+) -> Result<(), Errno> {
+    let mode = kind.bits() | perm.bits();
+
+    if cfg!(target_pointer_width = "64") {
+        // On 64-bit: dev_t fits in one 64-bit slot, cookies at arg4/arg5.
+        let res = path.with_nix_path(|cstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    SYS_mknodat,
+                    dirfd.as_fd().as_raw_fd(),
+                    cstr.as_ptr(),
+                    mode,
+                    dev,
+                    SYSCOOKIE_POOL.get(CookieIdx::MknodatArg4),
+                    SYSCOOKIE_POOL.get(CookieIdx::MknodatArg5),
+                )
+            }
+        })?;
+        Errno::result(res).map(drop)
+    } else if cfg!(target_pointer_width = "32") {
+        // On 32-bit: dev_t is 64-bit and consumes two 32-bit vararg slots.
+        // Split into low/high with correct endianness. Cookie at arg5 only.
+        let dev_low = (dev & 0xFFFF_FFFF) as libc::c_ulong;
+        let dev_high = (dev >> 32) as libc::c_ulong;
+
+        // Little-endian: low word first (arg3), high word second (arg4).
+        // Big-endian: high word first (arg3), low word second (arg4).
+        let (dev_arg3, dev_arg4) = if cfg!(target_endian = "little") {
+            (dev_low, dev_high)
+        } else if cfg!(target_endian = "big") {
+            (dev_high, dev_low)
+        } else {
+            unreachable!("BUG: unknown endian to split dev_t!");
+        };
+
+        let res = path.with_nix_path(|cstr| {
+            // SAFETY: In libc we trust.
+            unsafe {
+                syscall(
+                    SYS_mknodat,
+                    dirfd.as_fd().as_raw_fd(),
+                    cstr.as_ptr(),
+                    mode,
+                    dev_arg3,
+                    dev_arg4,
+                    SYSCOOKIE_POOL.get(CookieIdx::MknodatArg5),
+                )
+            }
+        })?;
+        Errno::result(res).map(drop)
+    } else {
+        unreachable!("BUG: unknown target pointer width!");
+    }
+}
+
 /// Safe fchdir(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_fchdir<Fd: AsFd>(dirfd: Fd) -> Result<(), Errno> {
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_fchdir,
+        syscall(
+            SYS_fchdir,
             dirfd.as_fd().as_raw_fd(),
             SYSCOOKIE_POOL.get(CookieIdx::FchdirArg1),
             SYSCOOKIE_POOL.get(CookieIdx::FchdirArg2),
@@ -973,6 +1123,147 @@ pub fn safe_fchdir<Fd: AsFd>(dirfd: Fd) -> Result<(), Errno> {
     .map(drop)
 }
 
+/// Safe faccessat2(2) confined by syscall cookies.
+///
+/// This calls faccessat2(2) with AT_EMPTY_PATH under the hood.
+pub fn safe_faccess<Fd: AsFd>(fd: Fd, mode: AccessFlags, mut flags: AtFlags) -> Result<(), Errno> {
+    // Remove AT_SYMLINK_NOFOLLOW and add AT_EMPTY_PATH to flags.
+    flags.remove(AtFlags::AT_SYMLINK_NOFOLLOW);
+    flags.insert(AtFlags::AT_EMPTY_PATH);
+
+    // SAFETY: No libc wrapper for faccessat2 yet.
+    Errno::result(unsafe {
+        syscall(
+            SYS_faccessat2,
+            fd.as_fd().as_raw_fd(),
+            empty_path() as *const c_char,
+            mode.bits(),
+            flags.bits(),
+            SYSCOOKIE_POOL.get(CookieIdx::Faccessat2Arg4),
+            SYSCOOKIE_POOL.get(CookieIdx::Faccessat2Arg5),
+        )
+    })
+    .map(drop)
+}
+
+/// Safe execveat(2) for executability check confined by syscall cookie.
+///
+/// This uses AT_EXECVE_CHECK | AT_EMPTY_PATH to check if the file
+/// referenced by fd is executable, without actually executing it.
+/// Requires Linux >= 6.14.
+// No HAVE_AT_EXECVE_CHECK check happens here, see syd::fd::check_executable.
+#[inline(always)]
+pub fn safe_execve_check<Fd: AsFd>(fd: Fd) -> Result<(), Errno> {
+    let flags = (AT_EXECVE_CHECK | AtFlags::AT_EMPTY_PATH).bits();
+
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        syscall(
+            SYS_execveat,
+            fd.as_fd().as_raw_fd(),
+            empty_path() as *const c_char,
+            empty_argv() as *const *const c_char,
+            empty_envp() as *const *const c_char,
+            flags,
+            SYSCOOKIE_POOL.get(CookieIdx::ExecveatArg5),
+        )
+    })
+    .map(drop)
+}
+
+/// Safe fchmodat(2) confined by syscall cookies.
+///
+/// Note: fchmodat(2) does not have a flags argument and always follows symlinks.
+#[inline(always)]
+pub(crate) fn safe_fchmodat<Fd: AsFd, P: NixPath + ?Sized>(
+    dirfd: Fd,
+    path: &P,
+    mode: Mode,
+) -> Result<(), Errno> {
+    let res = path.with_nix_path(|cstr| {
+        // SAFETY: In libc we trust.
+        unsafe {
+            syscall(
+                SYS_fchmodat,
+                dirfd.as_fd().as_raw_fd(),
+                cstr.as_ptr(),
+                mode.bits(),
+                SYSCOOKIE_POOL.get(CookieIdx::FchmodatArg3),
+                SYSCOOKIE_POOL.get(CookieIdx::FchmodatArg4),
+                SYSCOOKIE_POOL.get(CookieIdx::FchmodatArg5),
+            )
+        }
+    })?;
+    Errno::result(res).map(drop)
+}
+
+/// Safe fchmod(2) confined by syscall cookies.
+///
+/// This variant uses AT_EMPTY_PATH with fchmodat2(2) for fd-based chmod.
+#[inline(always)]
+pub(crate) fn safe_fchmod<Fd: AsFd>(dirfd: Fd, mode: Mode) -> Result<(), Errno> {
+    if *SYS_FCHMODAT2 < 0 {
+        return Err(Errno::ENOSYS);
+    }
+
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        syscall(
+            *SYS_FCHMODAT2,
+            dirfd.as_fd().as_raw_fd(),
+            empty_path() as *const c_char,
+            mode.bits(),
+            AtFlags::AT_EMPTY_PATH.bits(),
+            SYSCOOKIE_POOL.get(CookieIdx::Fchmodat2Arg4),
+            SYSCOOKIE_POOL.get(CookieIdx::Fchmodat2Arg5),
+        )
+    })
+    .map(drop)
+}
+
+/// Safe fchown(2) confined by syscall cookies.
+///
+/// This calls fchownat(2) with AT_EMPTY_PATH under the hood.
+#[inline(always)]
+pub(crate) fn safe_fchown<Fd: AsFd>(
+    dirfd: Fd,
+    owner: Option<Uid>,
+    group: Option<Gid>,
+) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        let (uid, gid) = chown_raw_ids(owner, group);
+        syscall(
+            SYS_fchownat,
+            dirfd.as_fd().as_raw_fd(),
+            empty_path() as *const c_char,
+            uid,
+            gid,
+            AtFlags::AT_EMPTY_PATH.bits(),
+            SYSCOOKIE_POOL.get(CookieIdx::FchownatArg5),
+        )
+    })
+    .map(drop)
+}
+
+/// Computes the raw UID and GID values to pass to a `*chown` call.
+// The cast is not unnecessary on all platforms.
+// Borrowed from nix-0.31.0.
+#[allow(clippy::unnecessary_cast)]
+fn chown_raw_ids(owner: Option<Uid>, group: Option<Gid>) -> (uid_t, gid_t) {
+    // According to the POSIX specification, -1 is used to indicate that owner and group
+    // are not to be changed.  Since uid_t and gid_t are unsigned types, we have to wrap
+    // around to get -1.
+    let uid = owner
+        .map(Into::into)
+        .unwrap_or_else(|| (0 as uid_t).wrapping_sub(1));
+
+    let gid = group
+        .map(Into::into)
+        .unwrap_or_else(|| (0 as gid_t).wrapping_sub(1));
+    (uid, gid)
+}
+
 /// Safe uname(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_uname() -> Result<UtsName, Errno> {
@@ -980,8 +1271,8 @@ pub fn safe_uname() -> Result<UtsName, Errno> {
 
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_uname,
+        syscall(
+            SYS_uname,
             &raw mut name,
             SYSCOOKIE_POOL.get(CookieIdx::UnameArg1),
             SYSCOOKIE_POOL.get(CookieIdx::UnameArg2),
@@ -1001,8 +1292,8 @@ pub fn safe_umask(mode: Mode) -> Mode {
     #[expect(clippy::cast_possible_truncation)]
     #[expect(clippy::cast_sign_loss)]
     let prev = unsafe {
-        libc::syscall(
-            libc::SYS_umask,
+        syscall(
+            SYS_umask,
             mode.bits(),
             SYSCOOKIE_POOL.get(CookieIdx::UmaskArg1),
             SYSCOOKIE_POOL.get(CookieIdx::UmaskArg2),
@@ -1010,11 +1301,34 @@ pub fn safe_umask(mode: Mode) -> Mode {
             SYSCOOKIE_POOL.get(CookieIdx::UmaskArg4),
             SYSCOOKIE_POOL.get(CookieIdx::UmaskArg5),
         )
-    } as libc::mode_t;
+    } as mode_t;
     #[expect(clippy::disallowed_methods)]
     Mode::from_bits(prev).expect("[BUG] umask returned invalid Mode")
 }
 
+/// Safe utimensat(2)
+///
+/// 1. This uses AT_EMPTY_PATH for fd-based timestamp updates.
+/// 2. This is NOT confined by syscall cookies because handling
+///    utimensat, utimensat_time64 and all that 32/64-bit is
+///    terribly error-prone so we call into libc.
+#[inline(always)]
+pub(crate) fn safe_utimensat<Fd: AsFd>(
+    dirfd: Fd,
+    times: &[libc::timespec; 2],
+) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        utimensat(
+            dirfd.as_fd().as_raw_fd(),
+            empty_path() as *const c_char,
+            &raw const times[0],
+            AtFlags::AT_EMPTY_PATH.bits() as c_int,
+        )
+    })
+    .map(drop)
+}
+
 /// Safe pipe2(2) confined by syscall cookies.
 #[inline(always)]
 pub fn safe_pipe2(flags: OFlag) -> Result<(OwnedFd, OwnedFd), Errno> {
@@ -1022,8 +1336,8 @@ pub fn safe_pipe2(flags: OFlag) -> Result<(OwnedFd, OwnedFd), Errno> {
 
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_pipe2,
+        syscall(
+            SYS_pipe2,
             fds.as_mut_ptr(),
             flags.bits(),
             SYSCOOKIE_POOL.get(CookieIdx::Pipe2Arg2),
