@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/chdir.rs: chdir(2) and fchdir(2) handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -28,10 +28,10 @@ pub(crate) fn sys_chdir(request: UNotifyEventRequest) -> ScmpNotifResp {
         #[expect(clippy::disallowed_methods)]
         if let Some(typ) = path_args.0.as_ref().unwrap().typ.as_ref() {
             if !typ.is_dir() {
-                return Ok(request.fail_syscall(Errno::ENOTDIR));
+                return Err(Errno::ENOTDIR);
             }
         } else {
-            return Ok(request.fail_syscall(Errno::ENOENT));
+            return Err(Errno::ENOENT);
         }
 
         // SAFETY: This is vulnerable to TOCTTOU.
@@ -60,11 +60,11 @@ pub(crate) fn sys_fchdir(request: UNotifyEventRequest) -> ScmpNotifResp {
         if let Some(typ) = path.typ.as_ref() {
             if !typ.is_dir() {
                 // Deny non-directory with ENOTDIR.
-                return Ok(request.fail_syscall(Errno::ENOTDIR));
+                return Err(Errno::ENOTDIR);
             }
         } else {
             // No file type, file disappeared mid-way?
-            return Ok(request.fail_syscall(Errno::ENOTDIR));
+            return Err(Errno::ENOTDIR);
         }
 
         // SAFETY: This is vulnerable to TOCTTOU.
