@@ -2,19 +2,17 @@
 // Syd: rock-solid application kernel
 // src/kernel/fanotify.rs: fanotify_mark(2) handler
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
 use std::os::fd::{AsFd, AsRawFd, FromRawFd, IntoRawFd, RawFd};
 
 use libseccomp::{ScmpArch, ScmpNotifResp};
-use nix::{
-    errno::Errno,
-    sys::fanotify::{Fanotify, MarkFlags, MaskFlags},
-};
+use nix::errno::Errno;
 
 use crate::{
+    compat::{Fanotify, MarkFlags, MaskFlags},
     config::PROC_FILE,
     confine::scmp_arch_bits,
     kernel::syscall_path_handler,
@@ -58,6 +56,13 @@ pub(crate) fn sys_fanotify_mark(request: UNotifyEventRequest) -> ScmpNotifResp {
         None => return request.fail_syscall(Errno::EINVAL),
     };
 
+    // Validate the FANotify FD.
+    let notify_fd = match RawFd::try_from(req.data.args[0]) {
+        Ok(fd) if fd < 0 => return request.fail_syscall(Errno::EBADF),
+        Ok(fd) => fd,
+        Err(_) => return request.fail_syscall(Errno::EBADF),
+    };
+
     // fanotify(7) requires read access to the file or directory,
     // with the exception of symbolic links. However, due to
     // the `/proc' indirection in the handler we must not pass
@@ -121,12 +126,6 @@ pub(crate) fn sys_fanotify_mark(request: UNotifyEventRequest) -> ScmpNotifResp {
             // path argument to avoid symlink TOCTOU.
             let pfd = XPathBuf::from_self_fd(fd.as_raw_fd())?;
 
-            // Validate the FANotify FD.
-            let notify_fd = RawFd::try_from(req.data.args[0]).or(Err(Errno::EBADF))?;
-            if notify_fd < 0 {
-                return Err(Errno::EBADF);
-            }
-
             // Get the FANotify FD.
             let notify_fd = request.get_fd(notify_fd).map(|notify_fd| {
                 // SAFETY: pidfd_getfd(2) returns valid FD on success.
