@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/mem.rs: Memory syscall handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -166,11 +166,11 @@ fn syscall_mem_handler(
 
     // Get the file descriptor before access check.
     let fd = if check_exec || check_append_only {
-        #[expect(clippy::cast_possible_truncation)]
-        let remote_fd = req.data.args[4] as RawFd;
-        if remote_fd < 0 {
-            return request.fail_syscall(Errno::EBADF);
-        }
+        let remote_fd = match RawFd::try_from(req.data.args[4]) {
+            Ok(fd) if fd < 0 => return request.fail_syscall(Errno::EBADF),
+            Ok(fd) => fd,
+            Err(_) => return request.fail_syscall(Errno::EBADF),
+        };
 
         match request.get_fd(remote_fd) {
             Ok(fd) => Some(fd),
@@ -430,10 +430,11 @@ fn syscall_mem_handler(
         // (a) Exec and Memory sandboxing are both disabled.
         // (b) Exec granted access, Memory sandboxing is disabled.
         // The first candidate is safe as sandboxing is disabled,
-        // however (b) should theoretically suffer from VFS TOCTOU as
-        // the fd can change after the access check. However, our tests
-        // show this is not the case, see vfsmod_toctou_mmap integration
-        // test.
+        // however (b) suffers from VFS TOCTOU as the fd can change
+        // after the access check. This is why by default we hook
+        // into mmap{,2} with ptrace(2) and guard it with the
+        // TOCTOU-mitigator. mmap{,2} only ends up here with
+        // trace/allow_unsafe_ptrace:1.
         return unsafe { request.continue_syscall() };
     }
 
