@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/mknod.rs: mknod(2) and mknodat(2) handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -11,11 +11,11 @@ use std::os::fd::AsFd;
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
-    sys::stat::{mknodat, Mode, SFlag},
+    sys::stat::{Mode, SFlag},
 };
 
 use crate::{
-    cookie::safe_umask,
+    cookie::{safe_mknodat, safe_umask},
     kernel::{syscall_path_handler, to_mode, to_sflag},
     lookup::FsFlags,
     proc::proc_umask,
@@ -119,7 +119,7 @@ fn syscall_mknod_handler(
     // Note, POSIX ACLs may override this.
     safe_umask(mask);
 
-    mknodat(
+    safe_mknodat(
         path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?,
         path.base(),
         kind,
