@@ -1,7 +1,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/net/recvmsg.rs: recvmsg(2), recvmmsg(2), and recvmmsg_time64 handlers
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -21,6 +21,7 @@ use nix::{
         socket::{SockaddrLike, SockaddrStorage},
         time::TimeSpec,
     },
+    unistd::Pid,
 };
 use zeroize::Zeroizing;
 
@@ -32,7 +33,6 @@ use crate::{
     config::MAX_RW_COUNT,
     confine::{is_valid_ptr, scmp_arch_bits},
     fd::{get_nonblock, has_recv_timeout},
-    fs::peer_inode,
     kernel::net::to_msgflags,
     path::XPath,
     req::UNotifyEventRequest,
@@ -201,6 +201,7 @@ pub(crate) fn handle_recvmsg(
             flags.force_cloexec() || call_flags.contains(MsgFlags::MSG_CMSG_CLOEXEC);
         let rand_fd = flags.force_rand_fd();
 
+        let pid = Pid::this();
         for (hdr, data) in cmsgs {
             let ctrl_buf = if hdr.cmsg_level == libc::SOL_SOCKET
                 && hdr.cmsg_type == libc::SCM_RIGHTS
@@ -236,13 +237,9 @@ pub(crate) fn handle_recvmsg(
                 #[expect(clippy::cast_ptr_alignment)]
                 let mut uc: libc::ucred = unsafe { *(data.as_ptr() as *const libc::ucred) };
 
-                // Get socket inode, and lookup pid by inode in sandbox unix map.
-                if let Some(pid) = peer_inode(&fd)
-                    .ok()
-                    .and_then(|inode| request.get_unix(inode))
-                    .map(|unix| unix.pid.as_raw())
-                {
-                    uc.pid = pid;
+                // Fix PID in credentials as necessary.
+                if uc.pid == pid.as_raw() {
+                    uc.pid = request.fix_cred_pid(&fd).as_raw();
                 }
 
                 // Serialize possibly-updated credentials back into a payload buffer.
@@ -404,6 +401,7 @@ fn handle_recvmmsg_internal(
 
     // Bitness of the sandbox process.
     let req = request.scmpreq;
+    let pid = Pid::this();
     let is32 = scmp_arch_bits(req.data.arch) == 32;
 
     // Read the user mmsghdr array.
@@ -651,13 +649,9 @@ fn handle_recvmmsg_internal(
                     #[expect(clippy::cast_ptr_alignment)]
                     let mut uc: libc::ucred = unsafe { *(data.as_ptr() as *const libc::ucred) };
 
-                    // Get socket inode, and lookup pid by inode in sandbox unix map.
-                    if let Some(pid) = peer_inode(&fd)
-                        .ok()
-                        .and_then(|inode| request.get_unix(inode))
-                        .map(|unix| unix.pid.as_raw())
-                    {
-                        uc.pid = pid;
+                    // Fix PID in credentials as necessary.
+                    if uc.pid == pid.as_raw() {
+                        uc.pid = request.fix_cred_pid(&fd).as_raw();
                     }
 
                     // Serialize possibly-updated credentials back into a payload buffer.
