@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/ptrace/mod.rs: ptrace(2) hooks
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -119,25 +119,45 @@ pub(crate) fn handle_ptrace_sysenter(
             let my_sandbox =
                 SandboxGuard::Read(sandbox.read().unwrap_or_else(|err| err.into_inner()));
 
+            // Call the system call handler, and record the result.
             let res = sysenter_mmap(pid, &my_sandbox, info);
             let exe = my_sandbox.enabled(Capability::CAP_EXEC);
 
             drop(my_sandbox); // release the read lock.
 
-            match (res, exe) {
-                (Ok(()), false) => {
-                    // Exec sandboxing disabled, continue process.
-                    Err(Errno::ECANCELED)
-                }
-                (Ok(()), true) => {
+            match res {
+                Ok(true) if exe => {
                     // Record mmap(2) pid for syscall-exit exec sandbox check.
                     cache.add_mmap(pid);
                     Ok(()) // Stop at syscall-exit.
                 }
-                (Err(errno), _) => {
-                    // Record errno(3) to restore at syscall-exit.
-                    cache.add_error(pid, Some(errno));
-                    Ok(()) // Stop at syscall-exit.
+                Ok(_) => {
+                    // Exec sandboxing disabled, continue process.
+                    Err(Errno::ECANCELED)
+                }
+                Err(errno) => {
+                    // Set system call to -1 to skip the system call.
+                    // Write error value into the return register.
+                    if let Err(errno) = ptrace_skip_syscall(pid, info.arch, Some(errno)) {
+                        // SAFETY: Failed to set return value, terminate the process.
+                        if errno != Errno::ESRCH {
+                            let _ = kill(pid, Some(Signal::SIGKILL));
+                        }
+                        Err(Errno::ESRCH)
+                    } else if cfg!(any(
+                        target_arch = "mips",
+                        target_arch = "mips32r6",
+                        target_arch = "mips64",
+                        target_arch = "mips64r6",
+                        target_arch = "s390x"
+                    )) {
+                        // Skip to syscall-stop to write return value.
+                        cache.add_error(pid, Some(errno));
+                        Ok(())
+                    } else {
+                        // Continue process.
+                        Err(Errno::ECANCELED)
+                    }
                 }
             }
         }
