@@ -99,7 +99,6 @@ const PID_MAX_LIMIT: libc::pid_t = 1 << 22;
 /// - `queue`: true if the system call is sigqueueinfo(2) or rt_sigqueueinfo.
 ///
 /// - `ScmpNotifResp`: Response indicating the result of the syscall handling.
-#[expect(clippy::cognitive_complexity)]
 fn syscall_signal_handler(request: UNotifyEventRequest, syscall: SignalSyscall) -> ScmpNotifResp {
     syscall_handler!(request, |request: UNotifyEventRequest| {
         let req = request.scmpreq;
@@ -135,11 +134,7 @@ fn syscall_signal_handler(request: UNotifyEventRequest, syscall: SignalSyscall)
                     Errno::EINVAL
                 }
             } else {
-                if pid > 0 {
-                    Errno::ESRCH
-                } else {
-                    Errno::EINVAL
-                }
+                Errno::ESRCH
             });
         }
 
