@@ -162,6 +162,8 @@ pub mod log;
 pub mod lookup;
 /// magic symlink utilities
 pub mod magic;
+/// Namespace utilities
+pub mod namespace;
 /// Interface to Open File Description locks
 pub mod ofd;
 /// /proc and syd(2) nom parsers
@@ -172,6 +174,8 @@ pub mod path;
 pub mod proc;
 /// ptrace(2) utilities
 pub mod ptrace;
+/// PTY utilities
+pub mod pty;
 /// seccomp(2) notify request handling
 pub mod req;
 /// Utilities to handle restarting syscalls
@@ -230,7 +234,7 @@ pub(crate) mod pool;
 /// The low-level interface for linux namespaces (containers)
 pub mod unshare;
 
-use std::{ffi::OsStr, os::fd::AsRawFd};
+use std::ffi::OsStr;
 
 use lexis::ToName;
 use libseccomp::ScmpVersion;
@@ -240,7 +244,6 @@ use nix::{
     sys::{
         resource::{getrlimit, Resource},
         signal::{sigaction, signal, SaFlags, SigAction, SigHandler, SigSet, Signal},
-        socket::{socket, AddressFamily, SockFlag, SockType},
         utsname::uname,
     },
     unistd::{Gid, Group, Uid, User},
@@ -931,282 +934,6 @@ pub fn human_size(bytes: usize) -> String {
     format!("{:.2}{}", size, SIZES[i])
 }
 
-#[expect(clippy::unnecessary_cast)]
-const SIOCGIFINDEX: u64 = libc::SIOCGIFINDEX as u64;
-#[expect(clippy::unnecessary_cast)]
-const SIOCGIFFLAGS: u64 = libc::SIOCGIFFLAGS as u64;
-#[expect(clippy::unnecessary_cast)]
-const SIOCSIFFLAGS: u64 = libc::SIOCSIFFLAGS as u64;
-
-/// Functionally equivalent to "ifconfig lo up".
-///
-/// Returns loopback interface index.
-pub fn loopback_set_up() -> Result<i32, Errno> {
-    // Create a socket
-    let sock = socket(
-        AddressFamily::Inet,
-        SockType::Stream,
-        SockFlag::empty(),
-        None,
-    )?;
-
-    // Prepare the interface request
-    let mut ifreq = libc::ifreq {
-        #[expect(clippy::cast_possible_wrap)]
-        ifr_name: [
-            b'l' as libc::c_char,
-            b'o' as libc::c_char,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
-        ],
-        // SAFETY: Manually initialize ifr_ifru.
-        ifr_ifru: unsafe { std::mem::zeroed() },
-    };
-
-    // SAFETY: Request loopback network device index.
-    let loindex: i32 = unsafe {
-        let mut ifr_index: libc::ifreq = std::mem::zeroed();
-        ifr_index.ifr_name = ifreq.ifr_name;
-        Errno::result(libc::syscall(
-            libc::SYS_ioctl,
-            sock.as_raw_fd(),
-            SIOCGIFINDEX as libc::c_ulong,
-            &mut ifr_index,
-        ))?;
-        // HACK: ifr_ifru is a union but libc crate does not define ifru_ivalue,
-        // which is a libc::c_int, so here we refer to it with ifru_mtu which
-        // is the same type.
-        ifr_index.ifr_ifru.ifru_mtu
-    };
-
-    // Set BIGTCP to LOOPBACK_BIGTCP_MAX if available.
-    // Note, we _must_ do this before setting up the network device.
-    use crate::config::LOOPBACK_BIGTCP_MAX;
-    match loopback_set_bigtcp(loindex, LOOPBACK_BIGTCP_MAX) {
-        Ok(_) => {
-            info!("ctx": "loopback_set_bigtcp",
-                "msg": "loopback network device has BIGTCP set",
-                "max": LOOPBACK_BIGTCP_MAX);
-        }
-        Err(errno) => {
-            info!("ctx": "loopback_set_bigtcp",
-                "msg": format!("set BIGTCP for loopback network device error: {errno}"),
-                "err": errno as i32);
-        }
-    };
-
-    // SAFETY: Get the current flags.
-    Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_ioctl,
-            sock.as_raw_fd(),
-            SIOCGIFFLAGS as libc::c_ulong,
-            &mut ifreq,
-        )
-    })?;
-
-    // Modify the flags to bring up the interface.
-    //
-    // SAFETY: We're accessing the field of a union here.
-    #[expect(clippy::cast_possible_truncation)]
-    unsafe {
-        ifreq.ifr_ifru.ifru_flags |= (libc::IFF_UP | libc::IFF_RUNNING) as libc::c_short
-    };
-
-    // SAFETY: Set the new flags.
-    Errno::result(unsafe {
-        libc::syscall(
-            libc::SYS_ioctl,
-            sock.as_raw_fd(),
-            SIOCSIFFLAGS as libc::c_ulong,
-            &mut ifreq,
-        )
-    })?;
-
-    Ok(loindex)
-}
-
-// libc crate does not define struct nl from linux/rtnetlink.h.
-#[repr(C)]
-#[derive(Debug, Clone, Copy)]
-struct nlattr {
-    nla_len: u16,
-    nla_type: u16,
-}
-
-// libc crate does not define struct nlmsg from linux/rtnetlink.h.
-#[repr(C)]
-#[derive(Debug, Clone, Copy)]
-struct nlmsg {
-    hdr: libc::nlmsghdr,
-    info: ifinfomsg,
-    attrs: [u8; 64],
-}
-
-// libc crate does not define struct ifinfomsg from linux/rtnetlink.h yet.
-#[repr(C)]
-#[derive(Debug, Copy, Clone)]
-struct ifinfomsg {
-    family: u8,
-    pad: u8,
-    ifi_type: u16, // ARPHRD_*
-    index: i32,    // Interface index
-    flags: u32,    // IFF_* flags
-    change: u32,   // IFF_* change mask
-}
-
-// These values are based on the Linux kernel headers.
-const IFLA_GRO_IPV4_MAX_SIZE: libc::c_ushort = 0x40;
-const IFLA_GRO_MAX_SIZE: libc::c_ushort = 0x3a;
-const IFLA_GSO_IPV4_MAX_SIZE: libc::c_ushort = 0x3f;
-const IFLA_GSO_MAX_SIZE: libc::c_ushort = 0x29;
-
-/// Functionally equivalent to "ip link set dev $ifindex g{r,s}o_max_size $max_size",
-/// which sets BIGTCP if available, see: https://lwn.net/Articles/884104/
-///
-/// Requires loopback interface index as argument.
-pub fn loopback_set_bigtcp(ifindex: i32, max_size: u32) -> Result<(), Errno> {
-    // Set BIGTCP to max_size if available.
-    use netlink_sys::{constants::*, Socket, SocketAddr};
-
-    // SAFETY: create netlink socket using netlink_sys for NETLINK_ROUTE.
-    let mut sock = Socket::new(NETLINK_ROUTE)
-        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
-    sock.bind(&SocketAddr::new(0, 0))
-        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
-
-    // SAFETY: Zero initialize. netlink message.
-    let mut msg: nlmsg = unsafe { std::mem::zeroed() };
-
-    // SAFETY: Set up netlink header.
-    let nl_hdr = &mut msg.hdr;
-    #[expect(clippy::arithmetic_side_effects)]
-    #[expect(clippy::cast_possible_truncation)]
-    {
-        nl_hdr.nlmsg_len = (size_of::<libc::nlmsghdr>() + size_of::<ifinfomsg>()) as u32;
-        nl_hdr.nlmsg_type = libc::RTM_NEWLINK;
-        nl_hdr.nlmsg_flags = (libc::NLM_F_REQUEST | libc::NLM_F_ACK) as u16;
-        nl_hdr.nlmsg_seq = 1;
-        nl_hdr.nlmsg_pid = 0;
-    }
-
-    // SAFETY: Populate ifinfomsg.
-    let info = &mut msg.info;
-    #[expect(clippy::cast_possible_truncation)]
-    {
-        info.family = libc::AF_UNSPEC as u8;
-        info.index = ifindex;
-        info.change = u32::MAX;
-    }
-
-    let mut offset = 0;
-    for &kind in &[
-        IFLA_GRO_IPV4_MAX_SIZE,
-        IFLA_GRO_MAX_SIZE,
-        IFLA_GSO_IPV4_MAX_SIZE,
-        IFLA_GSO_MAX_SIZE,
-    ] {
-        // SAFETY: Set attribute header.
-        #[expect(clippy::cast_ptr_alignment)]
-        let attr_ptr = unsafe { msg.attrs.as_mut_ptr().add(offset) as *mut nlattr };
-
-        // SAFETY: Write attribute metadata.
-        #[expect(clippy::arithmetic_side_effects)]
-        #[expect(clippy::cast_possible_truncation)]
-        unsafe {
-            (*attr_ptr).nla_type = kind;
-            (*attr_ptr).nla_len = (size_of::<nlattr>() + size_of::<u32>()) as u16;
-        }
-
-        // SAFETY: Write u32 payload.
-        #[expect(clippy::cast_ptr_alignment)]
-        unsafe {
-            let ptr = (attr_ptr as *mut u8).add(size_of::<nlattr>()) as *mut u32;
-            *ptr = max_size;
-        }
-
-        #[expect(clippy::arithmetic_side_effects)]
-        #[expect(clippy::cast_lossless)]
-        #[expect(clippy::cast_sign_loss)]
-        {
-            // SAFETY: NLA_ALIGN ensures proper alignment for netlink attributes
-            // as required by Linux kernel ABI.
-            offset += unsafe { libc::NLA_ALIGN((*attr_ptr).nla_len as libc::c_int) } as usize;
-        }
-    }
-
-    #[expect(clippy::arithmetic_side_effects)]
-    #[expect(clippy::cast_possible_truncation)]
-    {
-        msg.hdr.nlmsg_len += offset as u32;
-    }
-
-    // SAFETY: Cast to byte slice for send.
-    let buf = unsafe {
-        std::slice::from_raw_parts(
-            std::ptr::addr_of!(msg) as *const u8,
-            msg.hdr.nlmsg_len as usize,
-        )
-    };
-    sock.send(buf, 0)
-        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
-
-    // Receive response.
-    let (buf, _) = sock
-        .recv_from_full()
-        .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(libc::ENOSYS)))?;
-
-    // Check response messages for error.
-    let mut offset = 0;
-    #[expect(clippy::arithmetic_side_effects)]
-    #[expect(clippy::cast_ptr_alignment)]
-    while offset + size_of::<libc::nlmsghdr>() <= buf.len() {
-        // SAFETY: We're reading a netlink message header from validated bounds.
-        let hdr = unsafe { &*(buf.as_ptr().add(offset) as *const libc::nlmsghdr) };
-
-        let len = hdr.nlmsg_len as usize;
-        if len < size_of::<libc::nlmsghdr>() || offset + len > buf.len() {
-            return Err(Errno::EINVAL);
-        }
-
-        // Check for error message.
-        #[expect(clippy::cast_possible_truncation)]
-        if hdr.nlmsg_type == libc::NLMSG_ERROR as libc::c_ushort
-            && len >= size_of::<libc::nlmsghdr>() + size_of::<libc::nlmsgerr>()
-        {
-            // SAFETY: Enough data to safely parse nlmsgerr.
-            let err = unsafe {
-                &*(buf.as_ptr().add(offset + size_of::<libc::nlmsghdr>()) as *const libc::nlmsgerr)
-            };
-            if err.error != 0 {
-                return Err(Errno::from_raw(-err.error));
-            }
-        }
-
-        #[expect(clippy::cast_possible_truncation)]
-        #[expect(clippy::cast_possible_wrap)]
-        #[expect(clippy::cast_sign_loss)]
-        {
-            // SAFETY: nlmsg_len is kernel-aligned; advance to next message.
-            offset += unsafe { libc::NLA_ALIGN(len as i32) as usize };
-        }
-    }
-
-    Ok(())
-}
-
 /// Parse the given string into a UID.
 /// 1. use getpwnam_r(3)
 /// 2. parse as integer
