@@ -20,9 +20,9 @@ use memchr::arch::all::is_prefix;
 use nix::errno::Errno;
 use nom::{
     branch::alt,
-    bytes::complete::{tag, tag_no_case, take_while1},
+    bytes::complete::{escaped_transform, is_not, tag, tag_no_case, take_while1},
     character::complete::{char, digit1, one_of},
-    combinator::{all_consuming, map, opt, recognize},
+    combinator::{all_consuming, map, opt, recognize, value},
     error::{Error, ErrorKind},
     multi::separated_list1,
     sequence::preceded,
@@ -205,6 +205,112 @@ impl NetlinkCmd {
     }
 }
 
+/// Parsed mask command.
+///
+/// Format: `<op><pattern>[:<mask_all>[:<mask_dir>]]`
+/// - `<op>`: one of '+', '-', '^'
+/// - `<pattern>`: path pattern
+/// - `<mask_all>`: optional mask for files
+/// - `<mask_dir>`: optional mask for directories
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub(crate) struct MaskCmd {
+    pub(crate) op: char,
+    pub(crate) pattern: String,
+    pub(crate) mask_all: Option<String>,
+    pub(crate) mask_dir: Option<String>,
+}
+
+/// Parse a mask command suffix (after "mask" prefix is stripped).
+///
+/// Escape sequences: `\:` -> `:`, `\\` -> `\`
+pub(crate) fn parse_mask_cmd(input: &str) -> Result<MaskCmd, Errno> {
+    // Parse a field with escape sequences: `\:` -> `:`, `\\` -> `\`
+    fn escaped_field(input: &str) -> IResult<&str, String> {
+        escaped_transform(
+            is_not(":\\"),
+            '\\',
+            alt((value(":", tag(":")), value("\\", tag("\\")))),
+        )
+        .parse(input)
+        .or_else(|_: nom::Err<Error<&str>>| Ok((input, String::new())))
+    }
+
+    // Parse clear operation: `^`
+    fn parse_clear(input: &str) -> IResult<&str, MaskCmd> {
+        map(char('^'), |op| MaskCmd {
+            op,
+            pattern: String::new(),
+            mask_all: None,
+            mask_dir: None,
+        })
+        .parse(input)
+    }
+
+    // Parse delete operation: `-<pattern>`
+    fn parse_del(input: &str) -> IResult<&str, MaskCmd> {
+        let (rem, (op, pat)) = (char('-'), take_while1(|_| true)).parse(input)?;
+        Ok((
+            rem,
+            MaskCmd {
+                op,
+                pattern: pat.to_string(),
+                mask_all: None,
+                mask_dir: None,
+            },
+        ))
+    }
+
+    // Parse add operation: `+<pattern>[:<mask_all>[:<mask_dir>]]`
+    fn parse_add(input: &str) -> IResult<&str, MaskCmd> {
+        let (rem, op) = char('+').parse(input)?;
+
+        // Pattern (required).
+        let (rem, pattern) = escaped_field(rem)?;
+        if pattern.is_empty() {
+            return Err(nom::Err::Error(Error::new(input, ErrorKind::Fail)));
+        }
+
+        // Optional mask_all.
+        let (rem, mask_all) = if let Some(after) = rem.strip_prefix(':') {
+            let (r, f) = escaped_field(after)?;
+            (r, Some(f))
+        } else {
+            (rem, None)
+        };
+
+        // Optional mask_dir.
+        let mask_dir = if let Some(after) = rem.strip_prefix(':') {
+            let (r, f) = escaped_field(after)?;
+            if !r.is_empty() {
+                return Err(nom::Err::Error(Error::new(input, ErrorKind::Fail)));
+            }
+            Some(f)
+        } else if !rem.is_empty() {
+            return Err(nom::Err::Error(Error::new(input, ErrorKind::Fail)));
+        } else {
+            None
+        };
+
+        Ok((
+            "",
+            MaskCmd {
+                op,
+                pattern,
+                mask_all,
+                mask_dir,
+            },
+        ))
+    }
+
+    match all_consuming(alt((parse_clear, parse_del, parse_add)))
+        .parse(input)
+        .finish()
+    {
+        Ok((_, cmd)) => Ok(cmd),
+        Err(_) => Err(Errno::EINVAL),
+    }
+}
+
 // Parse a "bind" command string, returning `BindCmd` or `Errno::EINVAL`.
 //
 // Accepts: `bind<mod><src>:<dst>(:<opt>)?`
@@ -2130,4 +2236,170 @@ mod tests {
         assert_eq!(parse_fs_cmd("allow+ext4"), Err(Errno::EINVAL));
         assert_eq!(parse_fs_cmd("fs+ext4"), Err(Errno::EINVAL));
     }
+
+    #[test]
+    fn test_parse_mask_cmd_clear_op() {
+        let cmd = parse_mask_cmd("^").unwrap();
+        assert_eq!(cmd.op, '^');
+        assert_eq!(cmd.pattern, "");
+        assert_eq!(cmd.mask_all, None);
+        assert_eq!(cmd.mask_dir, None);
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_clear_rejects_args() {
+        assert_eq!(parse_mask_cmd("^/"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("^/foo"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("^:"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_del_simple() {
+        let cmd = parse_mask_cmd("-/proc/cmdline").unwrap();
+        assert_eq!(cmd.op, '-');
+        assert_eq!(cmd.pattern, "/proc/cmdline");
+        assert_eq!(cmd.mask_all, None);
+        assert_eq!(cmd.mask_dir, None);
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_del_with_colons() {
+        // Delete takes entire rest as pattern (no colon splitting).
+        let cmd = parse_mask_cmd("-/foo:bar:baz").unwrap();
+        assert_eq!(cmd.pattern, "/foo:bar:baz");
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_del_empty() {
+        assert_eq!(parse_mask_cmd("-"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_pattern_only() {
+        let cmd = parse_mask_cmd("+/proc/**").unwrap();
+        assert_eq!(cmd.op, '+');
+        assert_eq!(cmd.pattern, "/proc/**");
+        assert_eq!(cmd.mask_all, None);
+        assert_eq!(cmd.mask_dir, None);
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_pattern_and_mask_all() {
+        let cmd = parse_mask_cmd("+/***:/dev/null").unwrap();
+        assert_eq!(cmd.pattern, "/***");
+        assert_eq!(cmd.mask_all, Some("/dev/null".to_string()));
+        assert_eq!(cmd.mask_dir, None);
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_all_three_fields() {
+        let cmd = parse_mask_cmd("+/***:/dev/null:/dev/zero").unwrap();
+        assert_eq!(cmd.pattern, "/***");
+        assert_eq!(cmd.mask_all, Some("/dev/null".to_string()));
+        assert_eq!(cmd.mask_dir, Some("/dev/zero".to_string()));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_empty_pattern() {
+        assert_eq!(parse_mask_cmd("+"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("+:"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("+::"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_empty_mask_all() {
+        // Trailing colon with empty mask_all is allowed.
+        let cmd = parse_mask_cmd("+/foo:").unwrap();
+        assert_eq!(cmd.pattern, "/foo");
+        assert_eq!(cmd.mask_all, Some("".to_string()));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_add_empty_mask_dir() {
+        let cmd = parse_mask_cmd("+/foo:/bar:").unwrap();
+        assert_eq!(cmd.pattern, "/foo");
+        assert_eq!(cmd.mask_all, Some("/bar".to_string()));
+        assert_eq!(cmd.mask_dir, Some("".to_string()));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_escaped_colon_in_pattern() {
+        let cmd = parse_mask_cmd("+/foo\\:bar").unwrap();
+        assert_eq!(cmd.pattern, "/foo:bar");
+        assert_eq!(cmd.mask_all, None);
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_escaped_colon_in_mask_all() {
+        let cmd = parse_mask_cmd("+/foo:/bar\\:baz").unwrap();
+        assert_eq!(cmd.pattern, "/foo");
+        assert_eq!(cmd.mask_all, Some("/bar:baz".to_string()));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_escaped_colon_in_mask_dir() {
+        let cmd = parse_mask_cmd("+/***:/dev/null:/d\\:ev/zero").unwrap();
+        assert_eq!(cmd.pattern, "/***");
+        assert_eq!(cmd.mask_all, Some("/dev/null".to_string()));
+        assert_eq!(cmd.mask_dir, Some("/d:ev/zero".to_string()));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_multiple_escaped_colons() {
+        let cmd = parse_mask_cmd("+/a\\:b\\:c").unwrap();
+        assert_eq!(cmd.pattern, "/a:b:c");
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_backslash_not_before_colon_or_backslash() {
+        // Backslash followed by anything other than `:` or `\` is invalid.
+        assert_eq!(parse_mask_cmd("+/foo\\nbar"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("+/foo\\a"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_escaped_backslash() {
+        // `\\` -> `\`
+        let cmd = parse_mask_cmd("+/foo\\\\bar").unwrap();
+        assert_eq!(cmd.pattern, "/foo\\bar");
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_trailing_backslash() {
+        // Trailing backslash without escape target is invalid.
+        assert_eq!(parse_mask_cmd("+/foo\\"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_double_backslash_at_end() {
+        // `\\` at end -> single `\`
+        let cmd = parse_mask_cmd("+/foo\\\\").unwrap();
+        assert_eq!(cmd.pattern, "/foo\\");
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_invalid_operator() {
+        assert_eq!(parse_mask_cmd("!/foo"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("=/foo"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd("*/foo"), Err(Errno::EINVAL));
+        assert_eq!(parse_mask_cmd(""), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_too_many_colons() {
+        // More than 2 colons => trailing garbage.
+        assert_eq!(parse_mask_cmd("+/a:/b:/c:/d"), Err(Errno::EINVAL));
+    }
+
+    #[test]
+    fn test_parse_mask_cmd_glob_patterns() {
+        let cmd = parse_mask_cmd("+/proc/*/cmdline").unwrap();
+        assert_eq!(cmd.pattern, "/proc/*/cmdline");
+
+        let cmd = parse_mask_cmd("+/*/*msg").unwrap();
+        assert_eq!(cmd.pattern, "/*/*msg");
+
+        let cmd = parse_mask_cmd("+/home/**").unwrap();
+        assert_eq!(cmd.pattern, "/home/**");
+    }
 }
