@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/syd.rs: Main entry point
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 // Proxying code is based in part upon socksns crate which is:
 //     Copyright (c) 2020 Steven Engler
 //     SPDX-License-Identifier: MIT
@@ -53,62 +53,56 @@
 use std::{
     env,
     env::VarError,
-    ffi::{OsStr, OsString},
+    ffi::OsString,
     fs::OpenOptions,
-    io::{stdout, BufWriter, Write},
-    net::{IpAddr, SocketAddrV4, SocketAddrV6},
+    io::{stdin, stdout, BufWriter, Write},
     os::{
-        fd::{AsFd, AsRawFd, BorrowedFd, IntoRawFd},
-        unix::{ffi::OsStrExt, fs::OpenOptionsExt, net::UnixStream, process::CommandExt},
+        fd::{AsFd, AsRawFd, BorrowedFd, IntoRawFd, OwnedFd},
+        unix::{ffi::OsStrExt, fs::OpenOptionsExt},
     },
-    path::Path,
-    process::{ExitCode, Stdio},
+    process::{exit, ExitCode},
     str::FromStr,
 };
 
 use data_encoding::HEXLOWER;
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
-use memchr::arch::all::is_equal;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    libc::setdomainname,
-    pty::{grantpt, unlockpt},
     sched::{unshare, CloneFlags},
     sys::{
         resource::Resource,
-        signal::Signal,
-        socket::{bind, socket, AddressFamily, SockFlag, SockType, SockaddrIn, SockaddrIn6},
-        stat::{umask, Mode},
         wait::{Id, WaitPidFlag},
     },
-    time::{clock_gettime, ClockId},
-    unistd::{getgid, getpid, getuid, isatty, sethostname, setsid, Pid},
+    unistd::{fchdir, getgid, getpid, getuid, isatty, Pid},
 };
-use sendfd::SendWithFd;
 use syd::{
     caps,
-    compat::{clone3, set_name, set_pdeathsig, waitid, CloneArgs, WaitStatus},
+    compat::{set_name, waitid, ResolveFlag, WaitStatus},
     config::*,
-    confine::{confine_rlimit_zero, confine_scmp_madvise, ExportMode},
+    confine::{confine_rlimit_zero, confine_scmp_madvise, confine_scmp_wx_syd, ExportMode},
     err::err2no,
     error,
-    fd::{closeexcept, pidfd_open, set_cloexec},
-    fs::{format_clone_flags, format_clone_names, openpt, openpts},
+    fd::{closeexcept, fdclone},
+    fs::{format_clone_flags, format_clone_names},
     hash::{get_at_random_hex, hash, HashAlgorithm},
     hook::Supervisor,
     ignore_signals, info,
     landlock_policy::LandlockPolicy,
     log::log_init,
-    loopback_set_up,
+    lookup::safe_open_path,
+    namespace::{
+        ns_setup_net, ns_setup_pid, ns_setup_time, ns_setup_tor, ns_setup_user, ns_setup_uts,
+    },
     path::XPathBuf,
-    proc::proc_write,
+    proc::proc_open,
+    pty::pty_setup,
     rng::duprand,
     sandbox::Sandbox,
     seal::ensure_sealed,
     set_sigpipe_dfl, syd_code_name, syd_info,
     syslog::LogLevel,
-    warn, IgnoreSignalOpts,
+    IgnoreSignalOpts,
 };
 
 // Set global allocator to GrapheneOS allocator.
@@ -116,6 +110,7 @@ use syd::{
     not(coverage),
     not(feature = "prof"),
     not(target_os = "android"),
+    not(target_arch = "riscv64"),
     target_page_size_4k,
     target_pointer_width = "64"
 ))]
@@ -361,9 +356,30 @@ syd::main! {
         false
     };
 
-    // Local options handled by this function.
-    let sh = env::var_os(ENV_SH).unwrap_or(OsString::from(SYD_SH));
+    // Determine default shell to execute.
+    let mut is_rbash_def = false;
+    #[expect(clippy::disallowed_methods)]
+    let sh: Vec<_> = match env::var(ENV_SH) {
+        Ok(val) => shell_words::split(&val),
+        Err(VarError::NotPresent) => {
+            is_rbash_def = true;
+            shell_words::split(SYD_SH)
+        }
+        Err(error) => {
+            error!("ctx": "parse_shell", "op": "get_environment",
+                "msg": format!("detected invalid unicode in {ENV_SH}: {error}"),
+                "tip": format!("unset {ENV_SH} environment variable"));
+            return Err(error.into());
+        }
+    }?.into_iter().map(OsString::from).collect();
+    if sh.is_empty() {
+        error!("ctx": "parse_shell", "op": "split_shell",
+            "msg": format!("detected empty {ENV_SH}"),
+            "tip": format!("unset {ENV_SH} environment variable"));
+        return Err(shell_words::ParseError.into());
+    }
 
+    // Local options handled by this function.
     while let Some(arg) = parser.next()? {
         match arg {
             /*
@@ -456,7 +472,7 @@ syd::main! {
                 // with how bash and dash behaves.
                 user_parse = true;
                 if cmd_argv.is_empty() {
-                    cmd_argv.push(sh.clone());
+                    cmd_argv.extend(sh.clone());
                     cmd_argv.push(OsString::from("-c"));
                     cmd_argv.push(parser.value()?);
                 }
@@ -546,31 +562,18 @@ syd::main! {
     }
 
     // Prepare the command to execute, which may be a login shell.
-    let argv0 = if !cmd_argv.is_empty() {
-        Some(cmd_argv.remove(0))
-    } else {
-        None
-    };
-
-    let argv0 = match (export.is_some(), argv0, is_login) {
-        (true, _, _) => OsString::from(OsStr::from_bytes(MAGIC_FILE)),
-        (false, Some(argv0), false) => argv0,
-        (false, None, false) | (false, _, true) => {
-            #[expect(clippy::disallowed_methods)]
-            if cmd_arg0.is_none() {
-                // Allow user to override with -a.
-                cmd_arg0 = Some(match Path::new(&sh).file_name() {
-                    None => OsString::from("-"),
-                    Some(name) => {
-                        let mut p = OsString::from("-");
-                        p.push(name);
-                        p
-                    }
-                });
-            }
-            sh
+    let mut is_rbash = env::var_os(ENV_CD).is_some();
+    if cmd_argv.is_empty() {
+        cmd_argv = sh;
+        if cmd_arg0.is_none() {
+            // Allow user to override with -a.
+            cmd_arg0 = Some(OsString::from("-"));
         }
-    };
+        if is_rbash_def && export.is_none() {
+            is_rbash = true;
+        }
+    }
+    let argv0 = cmd_argv.remove(0);
 
     // Ignore all signals except the following signals:
     // SIGALRM, SIGCHLD, SIGKILL, SIGSTOP.
@@ -580,19 +583,14 @@ syd::main! {
     if sandbox.flags.allow_unsafe_prlimit() {
         opts.insert(IgnoreSignalOpts::SkipIgnoreCoreDump);
     }
-    match ignore_signals(opts) {
-        Ok(()) => {
-            info!("ctx": "ignore_signals",
-                "opt": opts, "msg": "ignored all signals for signal safety");
-        }
-        Err(errno) => {
-            error!("ctx": "ignore_signals",
-                "opt": opts, "err": errno as i32,
-                "msg": format!("ignoring signals failed: {errno}"),
-                "tip": "check with SYD_LOG=debug and/or submit a bug report");
-            return Err(errno.into());
-        }
-    }
+    ignore_signals(opts).inspect_err(|errno| {
+        error!("ctx": "ignore_signals",
+            "opt": opts, "err": *errno as i32,
+            "msg": format!("ignore signals failed: {errno}"),
+            "tip": "check with SYD_LOG=debug and/or submit a bug report");
+    })?;
+    info!("ctx": "ignore_signals",
+        "opt": opts, "msg": "ignored all signals for signal safety");
 
     // SAFETY: We cannot support NEWPID without NEWNS.
     // ie, pid namespace must have its own private /proc.
@@ -603,120 +601,66 @@ syd::main! {
         sandbox.set_unshare_pid(true);
     }
 
+    let has_ns_user = sandbox.flags.unshare_user();
+    let has_pid_max = sandbox.flags.unshare_pid() && sandbox.has_pid() && sandbox.pid_max > 0;
+    let has_ns_time = sandbox.flags.unshare_time();
+
+    // Save original UID/GID to map inside new user namespace.
+    let (uid, gid) = if has_ns_user {
+        (Some(getuid()), Some(getgid()))
+    } else {
+        (None, None)
+    };
+
+    // Open /proc safely as necessary.
+    // unshare/user:1 -> Need to write UID/GID mappings.
+    // unshare/pid:1  -> Need to write pid_max sysctl.
+    // unshare/time:1 -> Need to write time namespace offsets.
+    let fd_proc = if is_rbash || has_ns_user || has_pid_max || has_ns_time {
+        let fd = proc_open().inspect_err(|errno| {
+            error!("ctx": "setup_namespaces", "op": "open_procfs",
+                "err": *errno as i32,
+                "msg": format!("open /proc filesystem failed: {errno}"),
+                "tip": "mount procfs on top of /proc directory");
+        })?;
+
+        Some(fd)
+    } else {
+        None
+    };
+
+    // Switch to safe directory for rbash.
+    #[expect(clippy::disallowed_methods)]
+    if is_rbash {
+        let mut pfd = XPathBuf::from_pid(Pid::this())?;
+        pfd.push(b"fdinfo");
+
+        let fd_proc = fd_proc.as_ref().unwrap();
+        safe_open_path(fd_proc, &pfd, OFlag::empty(), ResolveFlag::RESOLVE_NO_XDEV).inspect_err(|errno| {
+            error!("ctx": "setup_restricted_shell", "op": "open_procfs",
+                "err": *errno as i32,
+                "msg": format!("open /proc filesystem failed: {errno}"),
+                "tip": "mount procfs on top of /proc directory");
+        }).and_then(fchdir).inspect_err(|errno| {
+            error!("ctx": "setup_restricted_shell", "op": "chdir_procfs",
+                "err": *errno as i32,
+                "msg": format!("change dir to /proc filesystem failed: {errno}"),
+                "tip": "mount procfs on top of /proc directory");
+        })?;
+    }
+
+    // Set up PTY sandboxing.
     let pty_child = if sandbox.has_pty()
-        && isatty(std::io::stdin()).unwrap_or(false)
-        && isatty(std::io::stdout()).unwrap_or(false)
+        && isatty(stdin()).unwrap_or(false)
+        && isatty(stdout()).unwrap_or(false)
     {
-        // TIP to be used in logging.
-        const TIP: &str = "put syd-pty(1) in PATH, set CARGO_BIN_EXE_syd-pty, or set sandbox/pty:off";
-
-        // Create a PIDFd of this process and clear O_CLOEXEC.
-        // PIDFD_NONBLOCK is equivalent to O_NONBLOCK,
-        // we use the latter because bionic libc doesn't define former yet.
-        #[expect(clippy::cast_sign_loss)]
-        let pidfd = pidfd_open(getpid(), OFlag::O_NONBLOCK.bits() as u32)
-            .inspect_err(|errno| {
-                error!("ctx": "spawn_pty", "op": "pidfd_open",
-                    "msg": format!("syd-pty pidfd_open error: {errno}"),
-                    "tip": TIP, "err": *errno as i32);
-            })?;
-        set_cloexec(&pidfd, false)?;
-
-        // Open a pseudoterminal device and clear O_CLOEXEC.
-        let pty_m = openpt(OFlag::O_RDWR | OFlag::O_NOCTTY)
-            .inspect_err(|errno| {
-                error!("ctx": "spawn_pty", "op": "openpt",
-                    "msg": format!("syd-pty openpt error: {errno}"),
-                    "tip": TIP, "err": *errno as i32);
-            })?;
-        set_cloexec(&pty_m, false)?;
-
-        // Grant access to PTY and unlock.
-        grantpt(&pty_m)?;
-        unlockpt(&pty_m)?;
-
-        // Open peer device.
-        // We are going to pass this end to the sandbox process.
-        // This uses TIOCGPTPEER ioctl(2) so O_NOFOLLOW is not needed.
-        let pty_s = openpts(&pty_m, OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_CLOEXEC)
-            .inspect_err(|errno| {
-                error!("ctx": "spawn_pty", "op": "openpts",
-                    "msg": format!("syd-pty openpts error: {errno}"),
-                    "tip": TIP, "err": *errno as i32);
-            })?;
-
-        // Prepare environment of the syd-pty process.
-        // Filter the environment variables to only include the list below:
-        // 1. PATH
-        // 2. LD_LIBRARY_PATH
-        // 3. SYD_PTY_RULES
-        // We do not need to pass SYD_PTY_DEBUG because we use -d as needed.
-        let safe_env: &[&[u8]] = &[b"PATH", b"LD_LIBRARY_PATH", b"SYD_PTY_RULES"];
-
-        // Spawn syd-pty process, and pass PTY main end to it.
         let pty_debug = env::var_os("SYD_PTY_DEBUG").is_some();
-        let mut cmd = if let Some(cmd) = env::var_os("CARGO_BIN_EXE_syd-pty") {
-            std::process::Command::new(cmd)
-        } else {
-            std::process::Command::new("syd-pty")
-        };
-        cmd.stdin(Stdio::inherit());
-        cmd.stdout(Stdio::inherit());
-        cmd.env_clear();
-        cmd.envs(env::vars_os()
-            .filter(|(key, _)| {
-                    safe_env.iter().any(|env| is_equal(key.as_bytes(), env))
-            })
-        );
-        if pty_debug {
-            cmd.arg("-d");
-            cmd.stderr(Stdio::inherit());
-        } else {
-            cmd.stderr(Stdio::null());
-        }
-        let mut buf = itoa::Buffer::new();
-        cmd.arg("-p");
-        cmd.arg(buf.format(pidfd.as_raw_fd()));
-        cmd.arg("-i");
-        cmd.arg(buf.format(pty_m.as_raw_fd()));
-        if let Some(ws) = sandbox.pty_ws_x() {
-            cmd.arg("-x");
-            cmd.arg(buf.format(ws));
-        }
-        if let Some(ws) = sandbox.pty_ws_y() {
-            cmd.arg("-y");
-            cmd.arg(buf.format(ws));
-        }
-        // SAFETY:
-        // Set umask(2) to a sane value.
-        // To let syd-pty(1) restore terminal at exit:
-        // 1. Do not set parent-death-signal.
-        // 2. Do not call setsid(2).
-        unsafe { cmd.pre_exec(|| {
-            let _ = set_name(c"syd_pty");
-            umask(Mode::from_bits_truncate(0o777));
-            Ok(())
-        }) };
-        if let Err(error) = cmd.spawn() {
-            let errno = err2no(&error);
-            error!("ctx": "spawn_pty", "op": "spawn",
-                "msg": format!("syd-pty spawn error: {error}"),
-                "tip": TIP, "err": errno as i32);
-            return Err(error.into());
-        }
-        drop(pidfd);
-        drop(pty_m);
-        if pty_debug {
-            warn!("ctx": "spawn_pty", "op": "forward_tty",
-                "msg": "syd-pty is now forwarding terminal I/O");
-        } else {
-            info!("ctx": "spawn_pty", "op": "forward_tty",
-                "msg": "syd-pty is now forwarding terminal I/O");
-        }
+        let pty_child = pty_setup(sandbox.pty_ws_x(), sandbox.pty_ws_y(), pty_debug)?;
 
-        // Pass the other end of the PTY pair to the sandbox process.
-        env::set_var(ENV_PTY_FD, buf.format(pty_s.as_raw_fd()));
-        Some(pty_s)
+        let mut buf = itoa::Buffer::new();
+        env::set_var(ENV_PTY_FD, buf.format(pty_child.as_raw_fd()));
+
+        Some(pty_child)
     } else {
         env::remove_var(ENV_PTY_FD);
         None
@@ -724,108 +668,16 @@ syd::main! {
 
     let proxy_debug = env::var_os("SYD_TOR_DEBUG").is_some();
     let proxy = if sandbox.has_proxy() {
-        // TIP to be used in logging.
-        const TIP: &str = "put syd-tor(1) in PATH, set CARGO_BIN_EXE_syd-tor, or set sandbox/proxy:off";
-
-        // sandbox/proxy:on implies unshare/net:1
+        // sandbox/proxy:on implies unshare/net:1.
         sandbox.set_unshare_net(true);
 
-        // Create a PIDFd of this process.
-        // PIDFD_NONBLOCK is equivalent to O_NONBLOCK,
-        // we use the latter because bionic libc doesn't define former yet.
-        #[expect(clippy::cast_sign_loss)]
-        let pidfd = pidfd_open(getpid(), OFlag::O_NONBLOCK.bits() as u32)
-            .inspect_err(|errno| {
-                error!("ctx": "spawn_tor", "op": "pidfd_open",
-                    "msg": format!("syd-tor pidfd_open error: {errno}"),
-                    "tip": TIP, "err": *errno as i32);
-            })?;
-
-        // Create a UNIX socket pair.
-        let (stream_parent, stream_child) = UnixStream::pair()
-            .inspect_err(|error| {
-                error!("ctx": "spawn_tor", "op": "socketpair",
-                    "msg": format!("syd-tor socketpair error: {error}"),
-                    "tip": TIP, "err": err2no(error) as i32);
-            })?;
-
-        // Unset the CLOEXEC flags on the file descriptors.
-        // PIDFds and Rust sockets are automatically CLOEXEC.
-        set_cloexec(&pidfd, false)?;
-        set_cloexec(&stream_parent, false)?;
-
-        // Prepare environment of the syd-tor process.
-        // Filter the environment variables to only include the list below:
-        // 1. PATH
-        // 2. LD_LIBRARY_PATH
-        // 3. SYD_TOR_RULES
-        // We do not need to pass SYD_TOR_DEBUG because we use -d as needed.
-        let safe_env: &[&[u8]] = &[b"PATH", b"LD_LIBRARY_PATH", b"SYD_TOR_RULES"];
-
-        // Spawn syd-tor process outside the namespace.
-        // Pass one end of the socket-pair to it.
-        let mut cmd = if let Some(cmd) = env::var_os("CARGO_BIN_EXE_syd-tor") {
-            std::process::Command::new(cmd)
-        } else {
-            std::process::Command::new("syd-tor")
-        };
-        cmd.stdin(Stdio::inherit());
-        cmd.stdout(Stdio::inherit());
-        cmd.env_clear();
-        cmd.envs(env::vars_os()
-            .filter(|(key, _)| {
-                    safe_env.iter().any(|env| is_equal(key.as_bytes(), env))
-            })
-        );
-        if proxy_debug {
-            cmd.arg("-d");
-            cmd.stderr(Stdio::inherit());
-        } else {
-            cmd.stderr(Stdio::null());
-        }
-        let mut buf = itoa::Buffer::new();
-        cmd.arg("-p");
-        cmd.arg(buf.format(pidfd.as_raw_fd()));
-        cmd.arg("-i");
-        cmd.arg(buf.format(stream_parent.as_raw_fd()));
-        // proxy/ext/unix has precedence over proxy/ext/host.
-        if let Some(ref proxy_ext_unix) = sandbox.proxy_ext_unix {
-            cmd.arg("-u");
-            cmd.arg(proxy_ext_unix);
-        } else {
-            let addr = format!("{}:{}", sandbox.proxy_ext_addr, sandbox.proxy_ext_port);
-            cmd.arg("-o");
-            cmd.arg(addr);
-        }
-        // SAFETY:
-        // 1. Set parent death signal to SIGKILL.
-        // 2. Create a new session.
-        // 3. Set umask(2) to a sane value.
-        unsafe { cmd.pre_exec(|| {
-            let _ = set_name(c"syd_tor");
-            set_pdeathsig(Some(Signal::SIGKILL))?;
-            setsid()?;
-            umask(Mode::from_bits_truncate(0o777));
-            Ok(())
-        }) };
-        if let Err(error) = cmd.spawn() {
-            let errno = err2no(&error);
-            error!("ctx": "spawn_tor", "op": "spawn",
-                "msg": format!("syd-tor spawn error: {error}"),
-                "tip": TIP, "err": errno as i32);
-            return Err(error.into());
-        }
-        drop(pidfd);
-        if proxy_debug {
-            warn!("ctx": "spawn_tor", "op": "forward_net",
-                "msg": format!("syd-tor is now forwarding external traffic to {}", sandbox.proxy_repr()));
-        } else {
-            info!("ctx": "spawn_tor", "op": "forward_net",
-                "msg": format!("syd-tor is now forwarding external traffic to {}", sandbox.proxy_repr()));
-        }
-
-        // Pass the other end of the socket-pair to the new namespace.
-        Some((stream_child, sandbox.proxy_port))
+        // Set up syd-tor.
+        Some(ns_setup_tor(
+                sandbox.proxy_ext_addr,
+                sandbox.proxy_ext_port,
+                sandbox.proxy_ext_unix.as_deref(),
+                sandbox.proxy_repr().as_str(),
+                proxy_debug)?)
     } else {
         None
     };
@@ -837,6 +689,9 @@ syd::main! {
     // view of /proc.
     let namespaces = sandbox.flags.namespaces();
     if namespaces == 0 {
+        // Drop /proc fd which may be open due to is_rbash.
+        drop(fd_proc);
+
         // No namespace arguments passed, run normally.
         return match Supervisor::run(
             sandbox,
@@ -854,29 +709,8 @@ syd::main! {
         };
     }
 
-    let id_buf = if sandbox.flags.unshare_user() {
-        // create the UID and GID mappings.
-        let uid = getuid().as_raw();
-        let gid = getgid().as_raw();
-
-        let map_root = sandbox.flags.map_root();
-        let uid_buf = {
-            let mut buf = Vec::new();
-            write!(&mut buf, "{} {uid} 1", if map_root { 0 } else { uid })?;
-            buf
-        };
-        let gid_buf = {
-            let mut buf = Vec::new();
-            write!(&mut buf, "{} {gid} 1", if map_root { 0 } else { gid })?;
-            buf
-        };
-        Some((uid_buf, gid_buf))
-    } else {
-        None
-    };
-
     // Tell the kernel to keep the capabilities after the unshare call.
-    // This is important because unshare() can change the user
+    // This is important because unshare(2) can change the user
     // namespace, which often leads to a loss of capabilities.
     caps::securebits::set_keepcaps(true)?;
 
@@ -885,215 +719,78 @@ syd::main! {
     let clone_flags = CloneFlags::from_bits_retain(namespaces);
     let clone_names = format_clone_flags(clone_flags);
     let clone_types = format_clone_names(&clone_names);
-    match unshare(clone_flags) {
-        Ok(()) => {
-            info!("ctx": "unshare_namespaces",
-                "ns": clone_names,
-                "msg": format!("unshared into {clone_types}"));
-        }
-        Err(errno) => {
-            error!("ctx": "unshare_namespaces",
-                "ns": clone_names, "err": errno as i32,
-                "msg": format!("unshare into {clone_types} failed: {errno}"),
-                "tip": "check with SYD_LOG=debug and/or set `unshare/user:1'");
-            return Err(errno.into());
-        }
+    unshare(clone_flags).inspect_err(|errno| {
+        error!("ctx": "unshare_namespaces",
+            "ns": &clone_names, "err": *errno as i32,
+            "msg": format!("unshare into {clone_types} failed: {errno}"),
+            "tip": "check with SYD_LOG=debug and/or set `unshare/user:1'");
+    })?;
+    info!("ctx": "setup_namespaces", "op": "unshare", "ns": &clone_names,
+        "msg": format!("unshared into {clone_types}"));
+    drop(clone_names);
+    drop(clone_types);
+
+    // Set up user namespace.
+    #[expect(clippy::disallowed_methods)]
+    if has_ns_user {
+        let fd_proc = fd_proc.as_ref().unwrap();
+        ns_setup_user(fd_proc, uid.unwrap(), gid.unwrap(), sandbox.flags.map_root())?;
     }
 
-    // Write uid/gid map for user namespace.
-    let pid = Pid::this();
-    if let Some((uid_buf, gid_buf)) = id_buf {
-        // Write "deny" to /proc/self/setgroups before writing to gid_map.
-        proc_write(pid, b"setgroups", b"deny")?;
-        proc_write(pid, b"gid_map", &gid_buf)?;
-        proc_write(pid, b"uid_map", &uid_buf)?;
-
-        // Set inheritable mask and ambient caps to retain caps after execve(2).
-        caps::securebits::set_keepcaps(true)?;
-        let permitted_caps = caps::read(None, caps::CapSet::Permitted)?;
-        caps::set(None, caps::CapSet::Inheritable, permitted_caps)?;
-        // Set the same capabilities as ambient, if necessary.
-        for flag in permitted_caps {
-            let cap = flag.try_into()?;
-            caps::raise(None, caps::CapSet::Ambient, cap)?;
-        }
+    // Set up PID namespace.
+    #[expect(clippy::disallowed_methods)]
+    if has_pid_max {
+        let fd_proc = fd_proc.as_ref().unwrap();
+        ns_setup_pid(fd_proc, sandbox.pid_max)?;
     }
 
-    // Bring up loopback device for net namespace.
+    // Set up time namespace.
+    if has_ns_time {
+        #[expect(clippy::disallowed_methods)]
+        let fd_proc = fd_proc.as_ref().unwrap();
+        ns_setup_time(fd_proc, sandbox.boottime, sandbox.monotime)?;
+    }
+    drop(fd_proc); // drop /proc fd.
+
+    // Set up network namespace.
     if sandbox.flags.unshare_net() {
-        // Set up the loopback interface.
-        // Warn on errors and continue.
-        match loopback_set_up() {
-            Ok(loindex) => {
-                info!("ctx": "loopback_set_up",
-                    "msg": format!("loopback network device is up with index:{loindex:#x}"),
-                    "idx": loindex);
-            }
-            Err(errno) => {
-                error!("ctx": "loopback_set_up",
-                    "msg": format!("set up loopback network device error: {errno}"),
-                    "err": errno as i32);
-            }
-        };
-
-        // Handle proxy sandboxing.
-        // Warn on errors and continue.
-        if let Some((stream_child, proxy_port)) = proxy {
-            let proxy_addr = sandbox.proxy_addr;
-            let ipv = if proxy_addr.is_ipv6() { 6 } else { 4 };
-            let fml = if ipv == 6 {
-                AddressFamily::Inet6
-            } else {
-                AddressFamily::Inet
-            };
-            let lfd = socket(
-                fml,
-                SockType::Stream,
-                SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
-                None,
-            )?;
-            let ret = match proxy_addr {
-                IpAddr::V4(addr_v4) => {
-                    let sockaddr = SockaddrIn::from(SocketAddrV4::new(addr_v4, proxy_port));
-                    bind(lfd.as_raw_fd(), &sockaddr)
-                }
-                IpAddr::V6(addr_v6) => {
-                    let sockaddr = SockaddrIn6::from(SocketAddrV6::new(addr_v6, proxy_port, 0, 0));
-                    bind(lfd.as_raw_fd(), &sockaddr)
-                }
-            };
-            match ret {
-                Ok(()) => {
-                    if proxy_debug {
-                        warn!("ctx": "bind_proxy",
-                            "msg": format!("proxy is now listening incoming IPv{ipv} requests from {proxy_addr}!{proxy_port}"));
-                    } else {
-                        info!("ctx": "bind_proxy",
-                            "msg": format!("proxy is now listening incoming IPv{ipv} requests from {proxy_addr}!{proxy_port}"));
-                    }
-                    let buf = [0u8; 1];
-                    let fds = [lfd.as_raw_fd()];
-                    match stream_child.send_with_fd(&buf, &fds) {
-                        Ok(_) => {
-                            if proxy_debug {
-                                warn!("ctx": "send_proxy_fd",
-                                    "msg": format!("proxy fd {} sent to syd-tor, IPv{ipv} traffic forwarding is now started \\o/",
-                                        lfd.as_raw_fd()));
-                                warn!("ctx": "send_proxy_fd",
-                                    "syd": "ping",
-                                    "msg": "Change return success. Going and coming without error.");
-                            } else {
-                                info!("ctx": "send_proxy_fd",
-                                    "msg": format!("proxy fd {} sent to syd-tor, IPv{ipv} traffic forwarding is now started \\o/",
-                                        lfd.as_raw_fd()));
-                                info!("ctx": "send_proxy_fd", "syd": "ping", "msg": "Change return success. Going and coming without error.");
-                            }
-                        }
-                        Err(error) => {
-                            error!("ctx": "send_proxy_fd",
-                                "msg": format!("proxy fd {} send to syd-tor error: {}, traffic will not be forwarded",
-                                    lfd.as_raw_fd(), err2no(&error)),
-                                "err": err2no(&error) as i32);
-                        }
-                    }
-                }
-                Err(errno) => {
-                    error!("ctx": "bind_proxy",
-                        "msg": format!("proxy bind to IPv{ipv} {proxy_addr}!{proxy_port} error: {errno}"),
-                        "err": errno as i32);
-                }
-            }
-            drop(stream_child);
-        }
+        ns_setup_net(proxy.as_ref(), sandbox.proxy_addr, sandbox.proxy_port, proxy_debug)?;
     }
+    drop(proxy); // close syd-tor end.
 
-    // Set host and domain name for uts namespace as necessary.
+    // Set up UTS namespace.
     if sandbox.flags.unshare_uts() {
-        if let Some(domainname) = sandbox.domainname.as_ref() {
-            let domainname = domainname.as_c_str();
-            let domainsize = domainname.to_bytes().len(); // without NUL.
-            let domainname = domainname.as_ptr().cast();
-
-            // SAFETY: No setdomainname(2) wrapper in nix yet.
-            if let Err(errno) = Errno::result(unsafe { setdomainname(domainname, domainsize) }) {
-                error!("ctx": "uts", "op": "setdomainname",
-                    "msg": format!("set NIS/YP domain name failed: {errno}"),
-                    "tip": "configure `uts/domain:none'",
-                    "err": errno as i32);
-                return Err(errno.into());
-            }
-        }
-
-        if let Some(hostname) = sandbox.hostname.as_ref() {
-            let hostname = hostname.as_c_str();
-            let hostname = OsStr::from_bytes(hostname.to_bytes());
-            if let Err(errno) = sethostname(hostname) {
-                error!("ctx": "uts", "op": "sethostname",
-                    "msg": format!("set host name failed: {errno}"),
-                    "tip": "configure `uts/host:none'",
-                    "err": errno as i32);
-                return Err(errno.into());
-            }
-        }
-    }
-
-    if sandbox.flags.unshare_time() {
-        let boffset = if let Some(boffset) = sandbox.boottime {
-            boffset
-        } else {
-            let btime = clock_gettime(ClockId::CLOCK_BOOTTIME)?;
-            // Into is necessary on 32-bit.
-            #[expect(clippy::useless_conversion)]
-            btime.tv_sec().checked_neg().ok_or(Errno::EOVERFLOW)?.into()
-        };
-        let moffset = if let Some(moffset) = sandbox.monotime {
-            moffset
-        } else {
-            let mtime = clock_gettime(ClockId::CLOCK_MONOTONIC)?;
-            // Into is necessary on 32-bit.
-            #[expect(clippy::useless_conversion)]
-            mtime.tv_sec().checked_neg().ok_or(Errno::EOVERFLOW)?.into()
-        };
-
-        if boffset != 0 || moffset != 0 {
-            // <clock-id> <offset-secs> <offset-nanosecs>
-            let mut buf = itoa::Buffer::new();
-            let mut data = "monotonic ".to_string();
-            data.push_str(buf.format(moffset));
-            data.push_str(" 0\nboottime ");
-            data.push_str(buf.format(boffset));
-            data.push_str(" 0\n");
-
-            proc_write(pid, b"timens_offsets", data.as_bytes())?;
-            info!("ctx": "set_boot_time", "msg": "set boot time in time namespace",
-                "timens_offsets": [moffset, boffset]);
-        }
+        ns_setup_uts(sandbox.hostname.as_deref(), sandbox.domainname.as_deref())?;
     }
 
     // Disable Speculative Store Bypass mitigations
     // for trace/allow_unsafe_exec_speculative:1
     let ssb = sandbox.flags.allow_unsafe_exec_speculative();
 
-    // SAFETY: Use clone3 with CLONE_PIDFD to avoid pid recycling.
-    let mut pid_fd: libc::c_int = -1;
-    let mut clone_args = CloneArgs {
-        flags: libc::CLONE_PIDFD as u64,
-        pidfd: (&raw mut pid_fd) as u64,
-        exit_signal: libc::SIGCHLD as u64,
-        ..Default::default()
-    };
+    // Prepare stack for new Syd process.
+    #[expect(clippy::cast_possible_truncation)]
+    let mut stack = [0u8; MAIN_STACK_SIZE as usize];
 
-    let child = clone3(&mut clone_args)?;
+    struct SydChildInfo {
+        sandbox: Sandbox,
+        pty_child: Option<OwnedFd>,
+        argv0: OsString,
+        cmd_argv: Vec<OsString>,
+        cmd_arg0: Option<OsString>,
+    }
 
     // Run the supervisor inside child process.
     // Syd will be the first process in new namespaces.
-    if child.as_raw() == 0 {
+    extern "C" fn syd_child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
+        // SAFETY: arg is a valid SydChildInfo structure.
+        let child_info: Box<SydChildInfo> = unsafe { Box::from_raw(arg as *mut SydChildInfo) };
+
         let retval = match Supervisor::run(
-            sandbox,
-            pty_child,
-            &argv0,
-            cmd_argv,
-            cmd_arg0,
+            child_info.sandbox,
+            child_info.pty_child,
+            &child_info.argv0,
+            child_info.cmd_argv,
+            child_info.cmd_arg0,
         ) {
             Ok(retval) => i32::from(retval),
             Err(error) => {
@@ -1102,26 +799,43 @@ syd::main! {
                 errno as i32
             }
         };
-        std::process::exit(retval);
+        exit(retval);
     }
 
-    // Only parent ends up here and `child` is child pid.
-    //
-    // Ensure we release all unneeded resources.
-    drop(cmd_arg0);
-    drop(cmd_argv);
-    drop(pty_child);
-    drop(sandbox);
+    // Prepare information for the Syd child.
+    let child_info = Box::new(SydChildInfo {
+        sandbox,
+        pty_child,
+        argv0,
+        cmd_argv,
+        cmd_arg0,
+    });
+    let child_info_ptr: *mut libc::c_void = Box::into_raw(child_info) as *mut libc::c_void;
+
+    // SAFETY: Use clone(2) with CLONE_PIDFD to avoid pid recycling.
+    let result = unsafe { fdclone(
+            syd_child_after_clone,
+            &mut stack,
+            child_info_ptr,
+            CloneFlags::empty(),
+            Some(libc::SIGCHLD),
+        )};
+
+    // SAFETY: Reconstruct and drop parent's copy.
+    drop(unsafe { Box::from_raw(child_info_ptr as *mut SydChildInfo) });
+
+    let (pid_fd, _) = result?;
 
     // SAFETY: Randomize the pid FD for hardening.
-    let pid_fd = duprand(pid_fd, OFlag::O_CLOEXEC | OFlag::O_EXCL)?;
+    let pid_fd_rand = duprand(pid_fd.as_raw_fd(), OFlag::O_CLOEXEC)?;
+    drop(pid_fd);
 
     // SAFETY: duprand returns a valid FD on success.
     // Create a BorrowedFd and NOT and OwnedFd, because this fd will
     // never be explicitly closed, it will be closed on exit. Moreover,
     // attempts to close any fd will fail due to the seccomp(2) filter
     // which will be applied before wait loop.
-    let pid_fd = unsafe { BorrowedFd::borrow_raw(pid_fd.into_raw_fd()) };
+    let pid_fd = unsafe { BorrowedFd::borrow_raw(pid_fd_rand.into_raw_fd()) };
 
     // Close all file descriptors but stderr and pidfd.
     // Set must be sorted because pidfd may be lower than stderr.
@@ -1231,15 +945,7 @@ syd::main! {
     }
 
     // Prevent executable memory.
-    const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
-    for name in ["mmap", "mmap2", "mprotect"] {
-        #[expect(clippy::disallowed_methods)]
-        ctx.add_rule_conditional(
-            ScmpAction::Allow,
-            ScmpSyscall::from_name(name).unwrap(),
-            &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
-        )?;
-    }
+    confine_scmp_wx_syd(&mut ctx)?;
 
     // Load the seccomp(2) filter.
     ctx.load()?;
