@@ -484,16 +484,20 @@ macro_rules! skip_if_root {
 }
 
 #[macro_export]
-macro_rules! skip_if_zfs {
-    () => {{
+macro_rules! skip_if_fs {
+    ($($fs:expr),* $(,)?) => {{
         let cwd = open(".", OFlag::O_PATH | OFlag::O_DIRECTORY, Mode::empty()).unwrap();
-        if syd::compat::FsType::get(cwd).unwrap().is_zfs() {
-            eprintln!("Test does not run correctly on ZFS, skipping!");
-            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
-            return Ok(());
-        } else {
-            eprintln!("Test does not run on ZFS, continuing...");
-        }
+        let current_fs = syd::compat::FsType::get(cwd).unwrap();
+        $(
+            if let Ok(fs_types) = syd::compat::FsType::from_name($fs) {
+                if fs_types.contains(&current_fs) {
+                    eprintln!("Test does not run correctly on {}, skipping!", $fs);
+                    std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+                    return Ok(());
+                }
+            }
+        )*
+        eprintln!("Filesystem check passed, continuing...");
     }};
 }
 
@@ -716,6 +720,14 @@ macro_rules! skip_unless_unshare {
             return Ok(());
         }
 
+        // Apparmor does not allow use of unprivileged namespaces by default.
+        if syd::confine::apparmor_enabled().unwrap_or(false) {
+            eprintln!("Apparmor detected!");
+            eprintln!("Test requires Linux namespaces ({:?}), skipping!", [$($ns),*]);
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        }
+
         match syd::confine::ns_enabled(flags) {
             Ok(false) => {
                 eprintln!("Test requires Linux namespaces ({:?}), skipping!", [$($ns),*]);
