@@ -32,10 +32,7 @@ mod child;
 mod config;
 mod error;
 mod ffi_util;
-mod idmap;
 mod linux;
-/// Utilities for Linux namespaces
-pub mod namespace;
 mod run;
 mod seccomp;
 mod status;
@@ -52,8 +49,6 @@ use nix::{errno::Errno, libc::pid_t};
 use crate::sandbox::RawIoctlMap;
 pub use crate::unshare::{
     error::Error,
-    idmap::{GidMap, UidMap},
-    namespace::Namespace,
     status::ExitStatus,
     stdio::{Fd, Stdio},
 };
