@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-cat.rs: Print out the rules of the given sandbox profile and exit
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -18,10 +18,17 @@ use syd::{
     sandbox::Sandbox,
 };
 
-// Set global allocator to mimalloc.
-#[cfg(all(not(feature = "prof"), target_pointer_width = "64"))]
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    not(target_arch = "riscv64"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
 #[global_allocator]
-static GLOBAL: mimalloc::MiMalloc = mimalloc::MiMalloc;
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
 
 // Set global allocator to tcmalloc if profiling is enabled.
 #[cfg(feature = "prof")]
@@ -154,9 +161,13 @@ fn help() {
 fn list() {
     println!("container");
     println!("core");
+    println!("cwd");
     println!("debug");
     println!("enforce");
     println!("firefox");
+    println!("fs");
+    println!("gui");
+    println!("hide");
     println!("immutable");
     println!("landlock");
     println!("lib");
@@ -165,8 +176,10 @@ fn list() {
     println!("nix");
     println!("noipv4");
     println!("noipv6");
+    println!("nomagic");
     println!("nomem");
     println!("nopie");
+    println!("noxdev");
     println!("oci");
     println!("paludis");
     println!("quiet");
@@ -289,6 +302,24 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "cwd" | "pwd" => {
+            println!("# Syd profile: CWD");
+            println!("# Number of rules: {}", syd::config::PROFILE_CWD.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_CWD {
+                println!("{command}");
+            }
+        }
+        "hide" => {
+            println!("# Syd profile: Hide");
+            println!("# Number of rules: {}", syd::config::PROFILE_HIDE.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_HIDE {
+                println!("{command}");
+            }
+        }
         "noipv4" => {
             println!("# Syd profile: NoIpv4");
             println!("# Number of rules: {}", syd::config::PROFILE_NOIPV4.len());
@@ -364,6 +395,24 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "nomagic" => {
+            println!("# Syd profile: Enforce No Magic Links");
+            println!("# Number of rules: {}", syd::config::PROFILE_NO_MAGIC.len());
+            println!("# Copyright (c) 2026 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_NO_MAGIC {
+                println!("{command}");
+            }
+        }
+        "noxdev" => {
+            println!("# Syd profile: Enforce No Cross Mounts");
+            println!("# Number of rules: {}", syd::config::PROFILE_NO_XDEV.len());
+            println!("# Copyright (c) 2026 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_NO_XDEV {
+                println!("{command}");
+            }
+        }
         "quiet" | "silent" => {
             println!("# Syd profile: Quiet");
             println!("# Number of rules: {}", syd::config::PROFILE_QUIET.len());
@@ -400,6 +449,24 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "fs" => {
+            println!("# Syd profile: Filesystem");
+            println!("# Number of rules: {}", syd::config::PROFILE_FS.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_FS {
+                println!("{command}");
+            }
+        }
+        "gui" => {
+            println!("# Syd profile: Graphical User Interface");
+            println!("# Number of rules: {}", syd::config::PROFILE_FS.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_GUI {
+                println!("{command}");
+            }
+        }
         "off" => {
             println!("# Syd profile: Off");
             println!("# Number of rules: {}", syd::config::PROFILE_OFF.len());
@@ -445,6 +512,18 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "nixstore" => {
+            println!("# Syd profile: Nix store");
+            println!(
+                "# Number of rules: {}",
+                syd::config::PROFILE_NIX_STORE.len()
+            );
+            println!("# Copyright (c) 2025 Emery Hemingway <emery+syd@slow.janky.email>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_NIX_STORE {
+                println!("{command}");
+            }
+        }
         _ => return Err(Errno::EINVAL.into()),
     }
 
