@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-fd.rs: Interact with remote file descriptors
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -10,7 +10,7 @@ use std::{
     env,
     ffi::OsString,
     os::{
-        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
+        fd::{AsRawFd, RawFd},
         unix::process::CommandExt,
     },
     process::{Command, ExitCode},
@@ -26,9 +26,8 @@ use nix::{
 use syd::{
     compat::getdents64,
     config::*,
-    fs::{
-        fd_status_flags, parse_fd, pidfd_getfd, pidfd_open, readlinkat, set_cloexec, PIDFD_THREAD,
-    },
+    fd::{fd_status_flags, parse_fd, pidfd_getfd, pidfd_open, set_cloexec, PIDFD_THREAD},
+    fs::readlinkat,
     path::{XPath, XPathBuf},
     rng::duprand,
 };
@@ -37,6 +36,8 @@ use syd::{
 #[cfg(all(
     not(coverage),
     not(feature = "prof"),
+    not(target_os = "android"),
+    not(target_arch = "riscv64"),
     target_page_size_4k,
     target_pointer_width = "64"
 ))]
@@ -172,8 +173,7 @@ syd::main! {
             Some(libc::AT_FDCWD) => {
                 let fd_rand = duprand(fd.as_raw_fd(), OFlag::empty())?;
                 drop(fd);
-                // SAFETY: duprand returns a valid FD on success.
-                unsafe { OwnedFd::from_raw_fd(fd_rand) }
+                fd_rand
             }
             Some(newfd) => {
                 // SAFETY: User should ensure no double-close happens.
