@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-pty.rs: Syd's PTY to STDIO bidirectional forwarder
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -12,7 +12,7 @@ use std::{
     process::{exit, ExitCode},
 };
 
-use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
+use libseccomp::{scmp_cmp, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     fcntl::{fcntl, splice, FcntlArg, OFlag, SpliceFFlags},
@@ -20,26 +20,47 @@ use nix::{
     sched::{unshare, CloneFlags},
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
-        prctl::set_dumpable,
+        resource::Resource,
         signal::{signal, sigprocmask, SigHandler, SigmaskHow, Signal},
         signalfd::{SfdFlags, SigSet, SignalFd},
-        termios::{cfmakeraw, tcgetattr, tcsetattr, OutputFlags, SetArg, Termios},
+        termios::{cfmakeraw, tcgetattr, tcsetattr, LocalFlags, OutputFlags, SetArg, Termios},
     },
-    unistd::{chdir, chroot, close, pipe2},
+    unistd::{chdir, chroot, pipe2},
 };
 use syd::{
-    compat::epoll_ctl_safe,
-    config::VDSO_SYSCALLS,
-    confine::{confine_mdwe, confine_scmp_madvise, extend_ioctl},
+    compat::{epoll_ctl_safe, set_dumpable},
+    config::{PTY_FCNTL_OPS, PTY_PRCTL_OPS, VDSO_SYSCALLS},
+    confine::{
+        confine_mdwe, confine_rlimit_zero, confine_scmp_fcntl, confine_scmp_madvise,
+        confine_scmp_prctl, confine_scmp_wx_syd, extend_ioctl, CLONE_NEWTIME,
+    },
     err::SydResult,
-    fs::{closeexcept, set_exclusive, set_nonblock, winsize_get, winsize_set},
+    fd::{close, closeexcept, set_exclusive, set_nonblock},
     ignore_signals,
-    ioctl::{TCGETS, TCGETS2, TCSETS, TCSETS2, TIOCGWINSZ, TIOCSWINSZ},
+    ioctl::IoctlMap,
     landlock_policy::LandlockPolicy,
+    pty::{winsize_get, winsize_set},
     rng::duprand,
     IgnoreSignalOpts,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    not(target_arch = "riscv64"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 // This is from <linux/tty.h>
 // libc does not export it...
 const N_TTY_BUF_SIZE: usize = 4096;
@@ -80,18 +101,12 @@ syd::main! {
     } = opts;
 
     // SAFETY: Randomize pidfd for hardening.
-    let fpid_fd = duprand(fpid.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fpid_fd = duprand(fpid.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(fpid);
     let fpid = fpid_fd;
 
     // SAFETY: Randomize PTY fd for hardening.
-    let fpty_fd = duprand(fpty.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fpty_fd = duprand(fpty.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(fpty);
     let fpty = fpty_fd;
 
@@ -99,10 +114,7 @@ syd::main! {
     let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;
 
     // SAFETY: Randomize the epoll fd for hardening.
-    let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(epoll);
     let epoll = Epoll(epoll_fd);
 
@@ -111,38 +123,20 @@ syd::main! {
     // SAFETY: Randomize pipe fds for hardening.
     let (pipe_pty_rd, pipe_pty_wr) = {
         let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
-        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
-        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
+        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC)?;
+        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC)?;
         (rd, wr)
     };
     let (pipe_std_rd, pipe_std_wr) = {
         let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
-        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
-        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
+        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC)?;
+        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC)?;
         (rd, wr)
     };
 
     // SAFETY: Randomize stdio(3) fds for hardening.
-    let fstd_rd = duprand(libc::STDIN_FILENO, OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
-    let fstd_wr = duprand(libc::STDOUT_FILENO, OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fstd_rd = duprand(libc::STDIN_FILENO, OFlag::O_CLOEXEC)?;
+    let fstd_wr = duprand(libc::STDOUT_FILENO, OFlag::O_CLOEXEC)?;
 
     // SAFETY: Set PTY to exclusive mode to harden against sniffing.
     set_exclusive(&fpty, true)?;
@@ -167,8 +161,8 @@ syd::main! {
     let fsig = {
         let fd = SignalFd::with_flags(&mask, SfdFlags::SFD_NONBLOCK | SfdFlags::SFD_CLOEXEC)?;
         duprand(fd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { SignalFd::from_owned_fd(OwnedFd::from_raw_fd(fd)) }
+            // SAFETY: dup(3) returned duplicate of a valid signal fd.
+            unsafe { SignalFd::from_owned_fd(fd) }
         })?
     };
 
@@ -178,7 +172,7 @@ syd::main! {
     confine(fsig.as_raw_fd(), debug, print)?;
 
     // Refresh terminal settings.
-    let tio = refresh_pty(&fstd_rd, &fpty);
+    let tio = refresh_pty(&fstd_rd, &fpty)?;
 
     // Refresh window size.
     refresh_win(&fstd_rd, &fpty, ws_x, ws_y);
@@ -203,9 +197,8 @@ syd::main! {
         &pipe_std_wr,
     );
 
-    if let Some(tio) = tio {
-        let _ = tcsetattr(&fstd_rd, SetArg::TCSANOW, &tio);
-    }
+    // Restore terminal settings at exit.
+    tcsetattr(&fstd_rd, SetArg::TCSANOW, &tio)?;
 
     Ok(match result {
         Ok(_) => ExitCode::SUCCESS,
@@ -438,23 +431,27 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     // For *2, we use hardcoded versions because not all libcs define them.
     // They are portable as `struct termios2` has the same size
     // across 32-bit and 64-bit architectures.
+    let arch = ScmpArch::native();
+    let ioctl = IoctlMap::new(None, true);
+    let names = [
+        "TCGETS",
+        "TCSETS",
+        "TCGETS2",
+        "TCSETS2",
+        "TIOCGWINSZ",
+        "TIOCSWINSZ",
+    ];
+    let mut iotty = Vec::with_capacity(names.len());
+    for name in names {
+        if let Some(req) = ioctl.get_value(name, arch) {
+            #[allow(clippy::unnecessary_cast)]
+            iotty.push(req as u64);
+        }
+    }
+
     #[expect(clippy::disallowed_methods)]
     let syscall = ScmpSyscall::from_name("ioctl").unwrap();
-    #[expect(clippy::unnecessary_cast)]
-    for request in [
-        *TCGETS,
-        *TCGETS2,
-        *TCSETS,
-        *TCSETS2,
-        *TIOCGWINSZ,
-        *TIOCSWINSZ,
-    ] {
-        let request = if let Some(request) = request {
-            request as u64
-        } else {
-            continue;
-        };
-
+    for request in iotty {
         ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
         if let Some(request) = extend_ioctl(request) {
             ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == request)])?;
@@ -462,35 +459,13 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     }
 
     // Allow safe fcntl(2) utility calls.
-    for name in ["fcntl", "fcntl64"] {
-        let syscall = match ScmpSyscall::from_name(name) {
-            Ok(syscall) => syscall,
-            Err(_) => continue,
-        };
-
-        for op in [libc::F_GETFD, libc::F_GETFL] {
-            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == op as u64)])?;
-        }
-    }
+    confine_scmp_fcntl(&mut ctx, PTY_FCNTL_OPS)?;
 
     // Allow safe prctl(2) operations.
-    let sysname = "prctl";
-    if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-        let op = libc::PR_SET_VMA as u64;
-        ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == op)])?;
-    }
+    confine_scmp_prctl(&mut ctx, PTY_PRCTL_OPS)?;
 
     // Prevent executable memory.
-    const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
-    for name in ["mmap", "mmap2", "mprotect"] {
-        #[expect(clippy::disallowed_methods)]
-        let syscall = ScmpSyscall::from_name(name).unwrap();
-        ctx.add_rule_conditional(
-            ScmpAction::Allow,
-            syscall,
-            &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
-        )?;
-    }
+    confine_scmp_wx_syd(&mut ctx)?;
 
     // We will ignore unshare errors next step and here we keep
     // with the expectation that we're inside the safe directory.
@@ -510,7 +485,7 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
             | CloneFlags::CLONE_NEWNS
             | CloneFlags::CLONE_NEWPID
             | CloneFlags::CLONE_NEWUTS
-            | CloneFlags::from_bits_retain(libc::CLONE_NEWTIME);
+            | CLONE_NEWTIME;
         if unshare(namespaces).is_ok() {
             chroot(".")?; // /proc/self/fdinfo.
             chdir("/")?; // prevent cwd leaking.
@@ -533,6 +508,19 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
 
         // Set the process dumpable attribute to not-dumpable.
         let _ = set_dumpable(false);
+
+        // Set nfiles, nprocs, and filesize rlimits to zero.
+        // Set locks, memory lock and msgqueue rlimits to zero.
+        // Set core dump file size to zero.
+        confine_rlimit_zero(&[
+            Resource::RLIMIT_CORE,
+            Resource::RLIMIT_FSIZE,
+            Resource::RLIMIT_NOFILE,
+            Resource::RLIMIT_NPROC,
+            Resource::RLIMIT_LOCKS,
+            Resource::RLIMIT_MEMLOCK,
+            Resource::RLIMIT_MSGQUEUE,
+        ])?;
     }
 
     if print_rules {
@@ -636,26 +624,22 @@ fn refresh_win<Fd1: AsFd, Fd2: AsFd>(
 }
 
 // Handle terminal settings.
-fn refresh_pty<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Option<Termios> {
-    let mut tio = if let Ok(tio) = tcgetattr(&src) {
-        tio
-    } else {
-        return None;
-    };
-
-    let orig_tio = tio.clone();
+fn refresh_pty<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Result<Termios, Errno> {
+    let mut tio = tcgetattr(&src)?;
 
     // Inherit host terminal settings for PTY.
-    let _ = tcsetattr(&dst, SetArg::TCSANOW, &tio);
+    tcsetattr(&dst, SetArg::TCSANOW, &tio)?;
 
     // Set raw mode for input TTY.
+    // Disable background processes from writing.
     // Enable output processing.
+    let orig_tio = tio.clone();
     cfmakeraw(&mut tio);
+    tio.local_flags.insert(LocalFlags::TOSTOP);
     tio.output_flags.insert(OutputFlags::OPOST);
+    tcsetattr(&src, SetArg::TCSANOW, &tio)?;
 
-    let _ = tcsetattr(&src, SetArg::TCSANOW, &tio);
-
-    Some(orig_tio)
+    Ok(orig_tio)
 }
 
 // Parse command line options.
