@@ -2,19 +2,21 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-sh.rs: confined shell
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
+#![allow(unused)]
+
 use std::{
     env,
     fs::File,
     io::{Read, Seek, SeekFrom, Stdin, Write},
     os::fd::{AsFd, AsRawFd, BorrowedFd},
     process::{exit, ExitCode},
-    time::Duration,
 };
 
+use dur::Duration;
 use linefeed::{Interface, ReadResult};
 use nix::{
     errno::Errno,
@@ -25,19 +27,20 @@ use syd::{
     config::*,
     cookie::safe_memfd_create,
     debug,
-    fs::{seal_memfd_all, set_cloexec},
+    fd::{seal_memfd_all, set_cloexec},
     get_user_home, get_user_name, human_size,
     io::ReadFd,
     lookup::safe_copy_if_exists,
     path::XPathBuf,
     syslog::LogLevel,
-    wordexp::*,
 };
 
 // Set global allocator to GrapheneOS allocator.
 #[cfg(all(
     not(coverage),
     not(feature = "prof"),
+    not(target_os = "android"),
+    not(target_arch = "riscv64"),
     target_page_size_4k,
     target_pointer_width = "64"
 ))]
@@ -49,11 +52,13 @@ static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc
 #[global_allocator]
 static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
 
+#[cfg(not(target_os = "android"))]
 enum Input {
     File(File),
     Stdin(Stdin),
 }
 
+#[cfg(not(target_os = "android"))]
 impl Read for Input {
     fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
         match self {
@@ -63,6 +68,7 @@ impl Read for Input {
     }
 }
 
+#[cfg(not(target_os = "android"))]
 impl AsFd for Input {
     fn as_fd(&self) -> BorrowedFd<'_> {
         match self {
@@ -72,10 +78,19 @@ impl AsFd for Input {
     }
 }
 
+#[cfg(not(target_os = "android"))]
 impl ReadFd for Input {}
 
+#[cfg(target_os = "android")]
+fn main() {
+    eprintln!("syd-sh: bionic libc doesn't support wordexp(3)!");
+    std::process::exit(libc::ENOSYS);
+}
+
+#[cfg(not(target_os = "android"))]
 syd::main! {
     use lexopt::prelude::*;
+    use syd::wordexp::*;
 
     // Set SIGPIPE handler to default.
     syd::set_sigpipe_dfl()?;
@@ -408,6 +423,7 @@ syd::main! {
     Ok(ExitCode::SUCCESS)
 }
 
+#[cfg(not(target_os = "android"))]
 fn help() {
     println!("Usage:");
     println!("  syd-sh [-helsx] [--] [_command_file_ [argument...]]");
@@ -417,7 +433,10 @@ fn help() {
     println!("Given -c with an argument, evaluate and print the result.");
 }
 
+#[cfg(not(target_os = "android"))]
 fn wrde2str(err: i32) -> String {
+    use syd::wordexp::*;
+
     match err {
         0 => "".to_string(),
         128 => "ERR?".to_string(),
@@ -432,10 +451,12 @@ fn wrde2str(err: i32) -> String {
     }
 }
 
+#[cfg(not(target_os = "android"))]
 fn quote(input: &str) -> String {
     format!("'{}'", input.replace("'", "'\\''"))
 }
 
+#[cfg(not(target_os = "android"))]
 const SHELL_BUILTINS: &[&str] = &[
     ".", "alias", "cd", "export", "hash", "readonly", "set", "shift", "source", "umask", "unalias",
     "unset",
