@@ -3,7 +3,7 @@
 // src/utils/syd-sys.rs: Given a number, print the matching syscall name and exit.
 //                       Given a regex, print case-insensitively matching syscall names and exit.
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -26,7 +26,7 @@ use nix::{
 use syd::{
     config::proc_init_simple,
     confine::print_seccomp_architectures,
-    fs::unix_inodes,
+    fd::unix_inodes,
     ioctl::{Ioctl, IoctlMap},
     parsers::sandbox::str2u64,
     proc::proc_unix_inodes,
@@ -38,6 +38,7 @@ use syd::{
     not(coverage),
     not(feature = "prof"),
     not(target_os = "android"),
+    not(target_arch = "riscv64"),
     target_page_size_4k,
     target_pointer_width = "64"
 ))]
