@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-tty.rs: Print the controlling terminal of the given process.
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -19,6 +19,8 @@ use syd::proc::proc_tty;
 #[cfg(all(
     not(coverage),
     not(feature = "prof"),
+    not(target_os = "android"),
+    not(target_arch = "riscv64"),
     target_page_size_4k,
     target_pointer_width = "64"
 ))]
