@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/workers/ipc.rs: `syd_ipc' remote-configuration thread
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -43,9 +43,12 @@ use crate::{
     alert,
     compat::{epoll_ctl_mod_safe, epoll_ctl_safe, MsgFlags},
     config::*,
-    confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules},
+    confine::{
+        confine_scmp_fcntl, confine_scmp_madvise, confine_scmp_prctl, confine_scmp_write,
+        confine_scmp_wx_syd, scmp_add_setid_rules,
+    },
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
-    fs::{closeexcept, peer_cred},
+    fd::{closeexcept, peer_creds},
     hash::SydHashMap,
     info,
     landlock::Errata,
@@ -271,14 +274,11 @@ impl IpcWorker {
         )?;
 
         // SAFETY: Randomize to make fd reuse harder.
-        let sock_fd = duprand(sock.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
+        let sock_fd = duprand(sock.as_raw_fd(), OFlag::O_CLOEXEC).inspect(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
             let mut fd_str = itoa::Buffer::new();
-            let fd_str = fd_str.format(fd);
+            let fd_str = fd_str.format(fd.as_raw_fd());
             std::env::set_var(ENV_IPC_UNIX_FD, fd_str);
-
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
         })?;
         drop(sock);
         let sock = sock_fd;
@@ -298,14 +298,11 @@ impl IpcWorker {
         let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;
 
         // SAFETY: Randomize the epoll fd to make fd reuse harder.
-        let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
+        let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).inspect(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
             let mut fd_str = itoa::Buffer::new();
-            let fd_str = fd_str.format(fd);
+            let fd_str = fd_str.format(fd.as_raw_fd());
             std::env::set_var(ENV_IPC_POLL_FD, fd_str);
-
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
         })?;
         drop(epoll);
         let epoll = Epoll(epoll_fd);
@@ -426,49 +423,13 @@ impl IpcWorker {
         }
 
         // Allow safe fcntl(2) utility calls.
-        for sysname in ["fcntl", "fcntl64"] {
-            let syscall = match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => syscall,
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                    continue;
-                }
-            };
-
-            for op in IPC_FCNTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == *op)])?;
-            }
-        }
+        confine_scmp_fcntl(&mut ctx, IPC_FCNTL_OPS)?;
 
         // Allow safe prctl(2) operations.
-        let sysname = "prctl";
-        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            for (_, op) in IPC_PRCTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
-            }
-        } else {
-            info!("ctx": "confine", "op": "allow_ipc_syscall",
-                "msg": format!("invalid or unsupported syscall {sysname}"));
-        }
+        confine_scmp_prctl(&mut ctx, IPC_PRCTL_OPS)?;
 
         // Prevent executable memory.
-        const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
-        for sysname in ["mmap", "mmap2", "mprotect"] {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
+        confine_scmp_wx_syd(&mut ctx)?;
 
         // Allow writes to the log-fd.
         // No proc_pid_mem(5) access required here.
@@ -478,33 +439,12 @@ impl IpcWorker {
         confine_scmp_madvise(&mut ctx)?;
 
         // Allow safe system calls.
-        for sysname in IPC_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow futex system calls.
-        for sysname in FUTEX_SYSCALLS {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_ipc_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow getid system calls.
-        for sysname in GET_ID_SYSCALLS {
+        for sysname in IPC_SYSCALLS
+            .iter()
+            .chain(FUTEX_SYSCALLS)
+            .chain(GETID_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
@@ -549,9 +489,11 @@ impl IpcWorker {
     fn init(&mut self, notif_pipe: (RawFd, RawFd)) -> SydResult<()> {
         // SAFETY: We use exit_group(2) here to bail,
         // because this unsharing is a critical safety feature.
-        if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
+        if let Err(errno) =
+            unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES | CloneFlags::CLONE_SYSVSEM)
+        {
             alert!("ctx": "boot", "op": "unshare_ipc_thread",
-                "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES): {errno}"),
+                "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES|CLONE_SYSVSEM): {errno}"),
                 "err": errno as i32);
             std::process::exit(101);
         }
@@ -705,7 +647,7 @@ impl IpcWorker {
                     epoll_ctl_safe(&epoll.0, fd, Some(event))?;
 
                     // Create Connection state.
-                    let creds = peer_cred(&stream)?;
+                    let creds = peer_creds(&stream)?;
                     let mut conn = Connection::new(creds, stream);
 
                     // In prompt mode, send initial prompt.
